import numpy as np
import pandas as pd

from .planar_geometry import PlanarGeometry
from .utils import rotation_matrix, rotation_matrices


class SheetGeometry(PlanarGeometry):
    """Geometry definitions for 2D sheets in 3D
    """

    @classmethod
    def update_all(cls, sheet):
        """
        Updates the sheet geometry by updating:
        * the edge vector coordinates
        * the edge lengths
        * the face centroids
        * the normals to each edge associated face
        * the face areas
        * the vertices heights (depends on geometry)
        * the face volumes (depends on geometry)

        """
        cls.update_dcoords(sheet)
        cls.update_length(sheet)
        cls.update_centroid(sheet)
        cls.update_height(sheet)
        cls.update_normals(sheet)
        cls.update_areas(sheet)
        cls.update_perimeters(sheet)
        cls.update_vol(sheet)

    @staticmethod
    def update_normals(sheet):
        """
        Updates the face_df `coords` columns as the face's vertices
        center of mass.
        """
        coords = sheet.coords
        face_pos = sheet.edge_df[["f" + c for c in coords]].to_numpy()
        srce_pos = sheet.edge_df[["s" + c for c in coords]].to_numpy()
        r_ij = sheet.edge_df[["d" + c for c in coords]].to_numpy()
        r_ai = srce_pos - face_pos
        normals = np.cross(r_ai, r_ij)
        sheet.edge_df[sheet.ncoords] = normals

    @staticmethod
    def update_areas(sheet):
        """
        Updates the normal coordniate of each (srce, trgt, face) face.
        """
        sheet.edge_df["sub_area"] = (
            np.linalg.norm(sheet.edge_df[sheet.ncoords], axis=1) / 2
        )
        # face_normals = sheet.upcast_face(
        #     sheet.edge_df.groupby("face")[sheet.ncoords].mean()
        # )
        # edge_orient = np.sign(
        #     np.sum(sheet.edge_df[sheet.ncoords] * face_normals, axis=0)
        # )
        # sheet.edge_df["sub_area"] *= edge_orient
        sheet.face_df["area"] = sheet.sum_face(sheet.edge_df["sub_area"])

    @staticmethod
    def update_vol(sheet):
        """
        Note that this is an approximation of the sheet geometry
        module.

        """
        sheet.edge_df["sub_vol"] = (
            sheet.upcast_srce(sheet.vert_df["height"]) * sheet.edge_df["sub_area"]
        )
        sheet.face_df["vol"] = sheet.sum_face(sheet.edge_df["sub_vol"])

    @classmethod
    def update_height(cls, sheet):
        """
        Update the height of the sheet vertices, based on the geometry
        specified in the sheet settings:

        `sheet.settings['geometry']` can be set to

          - `cylindrical`: the vertex height is
             measured with respect to the distance to the the axis
             specified in sheet.settings['height_axis'] (e.g `z`)
          - `flat`: the vertex height is
             measured with respect to the position on the axis
             specified in sheet.settings['height_axis']
          - 'spherical': the vertex height is measured with respect to its
             distance to the coordinate system centers
          - 'rod': the vertex height is measured with respect to its
             distance to the coordinate height axis if between the focii, and
             from the closest focus otherwise. The focii positions are updated
             before the height update.
                 ---------------
               /                 \
               |  *          *   |
               \                 /
                 ---------------
        In all the cases, this distance is shifted by an amount of
        `sheet.vert_df['basal_shift']`
        """
        w = sheet.settings.get("height_axis", "z")
        geometry = sheet.settings.get("geometry", "cylindrical")
        u, v = (c for c in sheet.coords if c != w)
        if geometry == "cylindrical":
            sheet.vert_df["rho"] = np.hypot(sheet.vert_df[v], sheet.vert_df[u])

        elif geometry == "flat":
            sheet.vert_df["rho"] = sheet.vert_df[w]

        elif geometry == "spherical":
            sheet.vert_df["rho"] = np.linalg.norm(sheet.vert_df[sheet.coords], axis=1)
        elif geometry == "rod":

            a, b = sheet.settings["ab"]
            w0 = b - a
            sheet.vert_df["rho"] = np.linalg.norm(sheet.vert_df[[u, v]], axis=1)
            sheet.vert_df["left_tip"] = sheet.vert_df[w] < -w0
            sheet.vert_df["right_tip"] = sheet.vert_df[w] > w0
            l_mask = sheet.vert_df[sheet.vert_df["left_tip"] == 1].index
            r_mask = sheet.vert_df[sheet.vert_df["right_tip"] == 1].index

            sheet.vert_df.loc[l_mask, "rho"] = cls.dist_to_point(
                sheet.vert_df.loc[l_mask], [0, 0, -w0], [u, v, w]
            )
            sheet.vert_df.loc[r_mask, "rho"] = cls.dist_to_point(
                sheet.vert_df.loc[r_mask], [0, 0, w0], [u, v, w]
            )

        elif sheet.settings["geometry"] == "surfacic":
            sheet.vert_df["rho"] = 1.0

        sheet.vert_df["height"] = sheet.vert_df["rho"] - sheet.vert_df["basal_shift"]

        edge_height = sheet.upcast_srce(sheet.vert_df[["height", "rho"]])
        edge_height.set_index(sheet.edge_df["face"], append=True, inplace=True)
        sheet.face_df[["height", "rho"]] = edge_height.mean(level="face")

    @classmethod
    def reset_scafold(cls, sheet):
        """
        Re-centers and (in the case of a rod sheet) resets the
        a-b parameters and tip masks
        """

        w = sheet.settings["height_axis"]
        u, v = (c for c in sheet.coords if c != w)

        cls.center(sheet)
        if sheet.settings["geometry"] == "rod":
            rho = np.linalg.norm(sheet.vert_df[[u, v]], axis=1)
            a = np.percentile(rho, 95)
            b = np.percentile(np.abs(sheet.vert_df[w]), 95)
            sheet.settings["ab"] = [a, b]

    @staticmethod
    def face_rotation(sheet, face, psi=0):
        """Returns a 3D rotation matrix such that the face normal points
        in the z axis

        Parameters
        ----------
        sheet: a :class:Sheet object
        face: int,
          the index of the face on which to rotate the sheet
        psi: float,
          Optional angle giving the rotation along the new `z` axis

        Returns
        -------
        rotation: (3, 3) np.ndarray
          The rotation matrix

        """
        normal = sheet.edge_df[sheet.edge_df["face"] == face][sheet.ncoords].mean()
        normal = normal / np.linalg.norm(normal)
        n_xy = np.linalg.norm(normal[["nx", "ny"]])
        theta = -np.arctan2(n_xy, normal.nz)
        direction = [normal.ny, -normal.nx, 0]
        r1 = rotation_matrix(theta, direction)
        if psi == 0:
            return r1
        else:
            return np.dot(rotation_matrix(psi, [0, 0, 1]), r1)

    @staticmethod
    def face_projected_pos(sheet, face, psi=0):
        """Returns the position of a face vertices projected on a plane
        perpendicular to the face normal, and translated so that the face
        center is at the center of the coordinate system


        Parameters
        ----------
        sheet: a :class:Sheet object
        face: int,
          the index of the face on which to rotate the sheet
        psi: float,
          Optional angle giving the rotation along the `z` axis

        Returns
        -------
        rot_pos: pd.DataFrame
           The rotated, relative positions of the face's vertices
        """

        face_orbit = sheet.edge_df[sheet.edge_df["face"] == face]["srce"]
        # n_sides = face_orbit.shape[0]
        # face_pos = np.repeat(
        #     sheet.face_df.loc[face, sheet.coords].values,
        #     n_sides).reshape(len(sheet.coords), n_sides).T
        rel_pos = (
            sheet.vert_df.loc[face_orbit.to_numpy(), sheet.coords].to_numpy()
            - sheet.face_df.loc[face, sheet.coords].to_numpy()
        )
        _, _, rotation = np.linalg.svd(rel_pos.astype(np.float), full_matrices=False)
        # rotation = cls.face_rotation(sheet, face, psi=psi)
        if psi != 0:
            rotation = np.dot(rotation_matrix(psi, [0, 0, 1]), rotation)
        rot_pos = pd.DataFrame(
            np.dot(rel_pos, rotation.T), index=face_orbit, columns=sheet.coords
        )
        return rot_pos

    @staticmethod
    def face_rotations(sheet):
        """Returns the (sheet.Ne, 3, 3) array of rotation matrices
        such that each rotation aligns the coordinate system along face normals

        """
        normals = sheet.edge_df.groupby("face")[sheet.ncoords].mean()
        normals = normals / np.linalg.norm(normals, axis=0)
        normals = sheet.upcast_face(normals)

        n_xy = np.linalg.norm(normals[["nx", "ny"]])
        theta = -np.arctan2(n_xy, normals.nz)

        direction = np.array(
            [normals.ny.to_numpy(), -normals.nx.to_numpy(), np.zeros(sheet.Ne)]
        ).T
        rots = rotation_matrices(theta, direction)

        return rots

    @classmethod
    def get_phis(cls, sheet):
        """Returns the 'latitude' of the vertices in the plane perpendicular
        to each face normal. For not-too-deformed faces, sorting vertices by this
        gives clockwize orientation.

        I think not-too-deformed means starconvex here.

        """
        rots = cls.face_rotations(sheet)

        rel_srce_pos = (
            sheet.edge_df[["sx", "sy", "sz"]]
            - sheet.edge_df[["fx", "fy", "fz"]].to_numpy()
        ).to_numpy()
        rotated = np.einsum("ikj, ik -> ij", rots, rel_srce_pos)
        return np.arctan2(rotated[:, 1], rotated[:, 0])

    @classmethod
    def sort_oriented_edges(cls, sheet):
        phis = cls.get_phis(sheet)
        sheet.edge_df = sheet.edge_df.sort_values(["face", phis]).reset_index()


class ClosedSheetGeometry(SheetGeometry):
    """Geometry for a closed 2.5D sheet.

    Apart from the geometry update from a normal sheet, the enclosed
    volume is also computed. The value is stored in `sheet.settings["lumen_vol"]`
    """

    @classmethod
    def update_all(cls, sheet):
        super().update_all(sheet)
        cls.update_lumen_vol(sheet)

    @staticmethod
    def update_lumen_vol(sheet):
        lumen_pos_faces = sheet.edge_df[["f" + c for c in sheet.coords]].to_numpy()
        lumen_sub_vol = (
            np.sum((lumen_pos_faces) * sheet.edge_df[sheet.ncoords].to_numpy(), axis=1)
            / 6
        )
        sheet.settings["lumen_vol"] = sum(lumen_sub_vol)


class EllipsoidGeometry(ClosedSheetGeometry):
    @staticmethod
    def update_height(eptm):

        a, b, c = eptm.settings["abc"]
        eptm.vert_df["theta"] = np.arcsin((eptm.vert_df.z / c).clip(-1, 1))
        eptm.vert_df["vitelline_rho"] = a * np.cos(eptm.vert_df["theta"])
        eptm.vert_df["basal_shift"] = (
            eptm.vert_df["vitelline_rho"] - eptm.specs["vert"]["basal_shift"]
        )
        eptm.vert_df["delta_rho"] = (
            np.linalg.norm(eptm.vert_df[["x", "y"]], axis=1)
            - eptm.vert_df["vitelline_rho"]
        ).clip(lower=0)

        SheetGeometry.update_height(eptm)

    @staticmethod
    def scale(eptm, scale, coords):
        SheetGeometry.scale(eptm, scale, coords)
        eptm.settings["abc"] = [u * scale for u in eptm.settings["abc"]]



"""Sheet behavior definitions
Each top level event is a series of actions leading to
the completion of an event.


"""

from .apoptosis_events import apoptosis
from .delamination_events import constriction
from .basic_events import division, type1_transition, face_elimination




# THIS FILE IS GENERATED FROM tyssue SETUP.PY
#
short_version = '0.6.7'
full_version = '0.6.7'
git_revision = 'fec8c8cc847f478c83656dffa850224e769b51db'
release = True
if release:
    version = full_version
else:
    version = short_version



import ipywidgets as ipw


def parameter_settings(eptm):
    specs = eptm.specs
    elements = []

    for element in ["edge", "vert", "face"]:
        if element not in specs:
            continue

        spec = specs[element]
        fts = []
        for param, val in spec.items():

            def update_param(change):
                specs[element][param] = change["new"]
                print(change)
                print("{} {} changed to {}".format(element, param, change["new"]))

            w = ipw.FloatText(val, description=param)
            w.observe(update_param, names="value")
            fts.append(w)
        elements.append(ipw.VBox(fts))

    return ipw.HBox(elements)



from ..json_parser import load_spec
import os

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def bulk_spec():
    """
    Default settings to perform edge subdivisions,
    see tyssue.io.point_cloud.py
    """
    specfile = os.path.join(CURRENT_DIR, "bulk.json")
    return load_spec(specfile)



"""
Matplotlib based plotting
"""
import shutil
import glob
import tempfile
import subprocess
import warnings
import pathlib


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


from matplotlib import cm
from matplotlib.path import Path
from matplotlib.patches import Polygon, FancyArrow, Arc, PathPatch
from matplotlib.collections import PatchCollection, PolyCollection

from ..config.draw import sheet_spec
from ..utils.utils import spec_updater, get_sub_eptm

COORDS = ["x", "y"]


def create_gif(history, output, num_frames=60, draw_func=None, margin=5, **draw_kwds):
    """Creates an animated gif of the recorded history.

    You need imagemagick on your system for this function to work.

    Parameters
    ----------

    history : a :class:`tyssue.History` object
    output : path to the output gif file
    num_frames : int, the number of frames in the gif
    draw_func : a drawing function
         this function must take a `sheet` object as first argument
         and return a `fig, ax` pair. Defaults to quick_edge_draw
         (aka sheet_view with quick mode)
    margin : int, the graph margins in percents, default 5
         if margin is -1, let the draw function decide

    **draw_kwds are passed to the drawing function

    """
    if draw_func is None:
        draw_func = sheet_view
        draw_kwds.update({"mode": "quick"})

    graph_dir = pathlib.Path(tempfile.mkdtemp())
    x, y = coords = draw_kwds.get("coords", history.sheet.coords[:2])
    sheet0 = history.retrieve(0)
    bounds = sheet0.vert_df[coords].describe().loc[["min", "max"]]
    delta = (bounds.loc["max"] - bounds.loc["min"]).max()
    margin = delta * margin / 100
    xlim = bounds.loc["min", x] - margin, bounds.loc["max", x] + margin
    ylim = bounds.loc["min", y] - margin, bounds.loc["max", y] + margin
    times = np.linspace(history.time_stamps[0], history.time_stamps[-1], num_frames)
    if len(history) < num_frames:
        for i, (t_, sheet) in enumerate(history):
            fig, ax = draw_func(sheet, **draw_kwds)
            if isinstance(ax, plt.Axes) and margin >= 0:
                ax.set(xlim=xlim, ylim=ylim)
            fig.savefig(graph_dir / f"sheet_{i:03d}")
            plt.close(fig)

            figs = glob.glob((graph_dir / "sheet_*.png").as_posix())
            figs.sort()

        for i, t in enumerate(times):
            index = np.where(history.time_stamps >= t)[0][0]
            fig = figs[index]
            shutil.copy(fig, graph_dir / f"movie_{i:04d}.png")
    else:
        for i, t in enumerate(times):
            sheet = history.retrieve(t)
            fig, ax = draw_func(sheet, **draw_kwds)
            if isinstance(ax, plt.Axes) and margin >= 0:
                ax.set(xlim=xlim, ylim=ylim)
            fig.savefig(graph_dir / f"movie_{i:04d}.png")
            plt.close(fig)

    try:
        proc = subprocess.run(
            ["convert", (graph_dir / "movie_*.png").as_posix(), output]
        )
    except Exception as e:
        print(
            "Converting didn't work, make sure imagemagick is available on your system"
        )
        raise e

    finally:
        shutil.rmtree(graph_dir)


def sheet_view(sheet, coords=COORDS, ax=None, **draw_specs_kw):
    """ Base view function, parametrizable
    through draw_secs

    The default sheet_spec specification is:

    {'edge': {
      'visible': True,
      'width': 0.5,
      'head_width': 0.2, # arrow head width for the edges
      'length_includes_head': True, # see matplotlib Arrow artist doc
      'shape': 'right',
      'color': '#2b5d0a', # can be an array
      'alpha': 0.8,
      'zorder': 1,
      'colormap': 'viridis'},
     'vert': {
      'visible': True,
      's': 100,
      'color': '#000a4b',
      'alpha': 0.3,
      'zorder': 2},
     'face': {
      'visible': False,
      'color': '#8aa678',
      'alpha': 1.0,
      'zorder': -1}
      }
    """
    draw_specs = sheet_spec()
    spec_updater(draw_specs, draw_specs_kw)
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    vert_spec = draw_specs["vert"]
    if vert_spec["visible"]:
        ax = draw_vert(sheet, coords, ax, **vert_spec)

    edge_spec = draw_specs["edge"]
    if edge_spec["visible"]:
        ax = draw_edge(sheet, coords, ax, **edge_spec)

    face_spec = draw_specs["face"]
    if face_spec["visible"]:
        ax = draw_face(sheet, coords, ax, **face_spec)

    ax.autoscale()
    ax.set_aspect("equal")
    return fig, ax


def draw_face(sheet, coords, ax, **draw_spec_kw):
    """Draws epithelial sheet polygonal faces in matplotlib
    Keyword values can be specified at the element
    level as columns of the sheet.face_df
    """

    draw_spec = sheet_spec()["face"]
    draw_spec.update(**draw_spec_kw)
    collection_specs = parse_face_specs(draw_spec, sheet)

    if "visible" in sheet.face_df.columns:
        edges = sheet.edge_df[sheet.upcast_face(sheet.face_df["visible"])].index
        _sheet = get_sub_eptm(sheet, edges)
        if _sheet is not None:
            sheet = _sheet
            color = collection_specs["facecolors"]
            if isinstance(color, np.ndarray):
                faces = sheet.face_df["face_o"].values.astype(np.uint32)
                collection_specs["facecolors"] = color.take(faces, axis=0)

    polys = sheet.face_polygons(coords)
    p = PolyCollection(polys, closed=True, **collection_specs)
    ax.add_collection(p)
    return ax


def parse_face_specs(face_draw_specs, sheet):

    collection_specs = {}
    color = face_draw_specs.get("color")

    if callable(color):
        color = color(sheet)
        face_draw_specs["color"] = color

    if color is None:
        return {}
    elif isinstance(color, str):
        collection_specs["facecolors"] = color
    elif hasattr(color, "__len__"):
        collection_specs["facecolors"] = _face_color_from_sequence(
            face_draw_specs, sheet
        )
    if "alpha" in face_draw_specs:
        collection_specs["alpha"] = face_draw_specs["alpha"]

    return collection_specs


def _face_color_from_sequence(face_spec, sheet):
    color_ = face_spec["color"]
    cmap = cm.get_cmap(face_spec.get("colormap", "viridis"))
    color_min, color_max = face_spec.get("color_range", (color_.min(), color_.max()))

    if color_.shape in [(sheet.Nf, 3), (sheet.Nf, 4)]:
        return color_

    elif color_.shape == (sheet.Nf,):
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((sheet.Nf, 3)) * 0.5

        normed = (color_ - color_min) / (color_max - color_min)
        return cmap(normed)

    else:
        raise ValueError(
            "shape of `face_spec['color']` must be either (Nf, 3), (Nf, 4) or (Nf,)"
        )


def draw_vert(sheet, coords, ax, **draw_spec_kw):
    """Draw junction vertices in matplotlib
    """
    draw_spec = sheet_spec()["vert"]
    draw_spec.update(**draw_spec_kw)

    x, y = coords
    if "z_coord" in sheet.vert_df.columns:
        pos = sheet.vert_df.sort_values("z_coord")[coords]
    else:
        pos = sheet.vert_df[coords]
    ax.scatter(pos[x], pos[y], **draw_spec_kw)
    return ax


def draw_edge(sheet, coords, ax, **draw_spec_kw):
    """
    """
    draw_spec = sheet_spec()["edge"]
    draw_spec.update(**draw_spec_kw)

    x, y = coords
    dx, dy = ("d" + c for c in coords)
    app_length = np.hypot(sheet.edge_df[dx], sheet.edge_df[dy])

    patches = []
    arrow_specs, collections_specs = parse_edge_specs(draw_spec, sheet)

    for idx, edge in sheet.edge_df[app_length > 1e-6].iterrows():
        srce = int(edge["srce"])
        arrow = FancyArrow(
            sheet.vert_df.loc[srce, x],
            sheet.vert_df.loc[srce, y],
            sheet.edge_df.loc[idx, dx],
            sheet.edge_df.loc[idx, dy],
            **arrow_specs,
        )
        patches.append(arrow)
    ax.add_collection(PatchCollection(patches, False, **collections_specs))
    return ax


def parse_edge_specs(edge_draw_specs, sheet):

    arrow_keys = ["head_width", "length_includes_head", "shape"]
    arrow_specs = {
        key: val for key, val in edge_draw_specs.items() if key in arrow_keys
    }
    collection_specs = {}
    if "color" in edge_draw_specs:
        if callable(edge_draw_specs["color"]):
            edge_draw_specs["color"] = edge_draw_specs["color"](sheet)

        if isinstance(edge_draw_specs["color"], str):
            collection_specs["edgecolors"] = edge_draw_specs["color"]
        elif hasattr(edge_draw_specs["color"], "__len__"):
            collection_specs["edgecolors"] = _wire_color_from_sequence(
                edge_draw_specs, sheet
            )

    if "width" in edge_draw_specs:
        collection_specs["linewidths"] = edge_draw_specs["width"]
    if "alpha" in edge_draw_specs:
        collection_specs["alpha"] = edge_draw_specs["alpha"]
    return arrow_specs, collection_specs


def _wire_color_from_sequence(edge_spec, sheet):
    """
    """
    color_ = edge_spec["color"]

    color_min, color_max = edge_spec.get("color_range", (color_.min(), color_.max()))
    cmap = cm.get_cmap(edge_spec.get("colormap", "viridis"))
    if color_.shape in [(sheet.Nv, 3), (sheet.Nv, 4)]:
        return (sheet.upcast_srce(color_) + sheet.upcast_trgt(color_)) / 2
    elif color_.shape == (sheet.Nv,):
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((sheet.Ne, 3)) * 0.7
        if not hasattr(color_, "index"):
            color_ = pd.Series(color_, index=sheet.vert_df.index)
        color_ = (sheet.upcast_srce(color_) + sheet.upcast_trgt(color_)) / 2
        return cmap((color_ - color_min) / (color_max - color_min))

    elif color_.shape in [(sheet.Ne, 3), (sheet.Ne, 4)]:
        return color_
    elif color_.shape == (sheet.Ne,):
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((sheet.Nv, 3)) * 0.7
        return cmap((color_ - color_min) / (color_max - color_min))


def quick_edge_draw(sheet, coords=["x", "y"], ax=None, **draw_spec_kw):

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    lines_x, lines_y = _get_lines(sheet, coords)
    ax.plot(lines_x, lines_y, **draw_spec_kw)
    ax.set_aspect("equal")
    return fig, ax


def _get_lines(sheet, coords):

    lines_x, lines_y = np.zeros(2 * sheet.Ne), np.zeros(2 * sheet.Ne)
    scoords = ["s" + c for c in coords]
    tcoords = ["t" + c for c in coords]
    if set(scoords + tcoords).issubset(sheet.edge_df.columns):
        srce_x, srce_y = sheet.edge_df[scoords].values.T
        trgt_x, trgt_y = sheet.edge_df[tcoords].values.T
    else:
        srce_x, srce_y = sheet.upcast_srce(sheet.vert_df[coords]).values.T
        trgt_x, trgt_y = sheet.upcast_trgt(sheet.vert_df[coords]).values.T

    lines_x[::2] = srce_x
    lines_x[1::2] = trgt_x
    lines_y[::2] = srce_y
    lines_y[1::2] = trgt_y
    # Trick from https://github.com/matplotlib/
    # matplotlib/blob/master/lib/matplotlib/tri/triplot.py#L65
    lines_x = np.insert(lines_x, slice(None, None, 2), np.nan)
    lines_y = np.insert(lines_y, slice(None, None, 2), np.nan)
    return lines_x, lines_y


def plot_forces(
    sheet, geom, model, coords, scaling, ax=None, approx_grad=None, **draw_specs_kw
):
    """Plot the net forces at each vertex, with their amplitudes multiplied
    by `scaling`. To be clear, this is the oposite of the gradient - grad E.
    """
    draw_specs = sheet_spec()
    spec_updater(draw_specs, draw_specs_kw)
    gcoords = ["g" + c for c in coords]
    if approx_grad is not None:
        app_grad = approx_grad(sheet, geom, model)
        grad_i = (
            pd.DataFrame(
                index=sheet.active_verts,
                data=app_grad.reshape((-1, len(sheet.coords))),
                columns=["g" + c for c in sheet.coords],
            )
            * scaling
        )

    else:
        grad_i = model.compute_gradient(sheet, components=False) * scaling

    arrows = pd.DataFrame(columns=coords + gcoords, index=sheet.vert_df.index)
    arrows[coords] = sheet.vert_df[coords]
    arrows[gcoords] = -grad_i[gcoords]  # F = -grad E

    if ax is None:
        fig, ax = quick_edge_draw(sheet, coords)
    else:
        fig = ax.get_figure()

    for _, arrow in arrows.iterrows():
        ax.arrow(*arrow, **draw_specs["grad"])
    return fig, ax


def plot_scaled_energies(sheet, geom, model, scales, ax=None):
    """Plot scaled energies

    Parameters
    ----------
    sheet: a:class: Sheet object
    geom: a :class:`Geometry` class
    model: a :class:'Model'
    scales: np.linspace of float

    Returns
    -------
    fig: a :class:matplotlib.figure.Figure instance
    ax: :class:matplotlib.Axes instance, default None
    """

    from ..utils import scaled_unscaled

    def get_energies():
        energies = np.array([e.mean() for e in model.compute_energy(sheet, True)])

        return energies

    energies = np.array(
        [scaled_unscaled(get_energies, scale, sheet, geom) for scale in scales]
    )
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    ax.plot(scales, energies.sum(axis=1), "k-", lw=4, alpha=0.3, label="total")
    for e, label in zip(energies.T, model.labels):
        ax.plot(scales, e, label=label)
    ax.legend()
    return fig, ax


def get_arc_data(sheet):

    srce_pos = sheet.upcast_srce(sheet.vert_df[sheet.coords])
    trgt_pos = sheet.upcast_trgt(sheet.vert_df[sheet.coords])

    radius = 1 / sheet.edge_df["curvature"]

    e_x = sheet.edge_df["dx"] / sheet.edge_df["length"]
    e_y = sheet.edge_df["dy"] / sheet.edge_df["length"]

    center_x = (srce_pos.x + trgt_pos.x) / 2 - e_y * (radius - sheet.edge_df["sagitta"])

    center_y = (srce_pos.y + trgt_pos.y) / 2 - e_x * (radius - sheet.edge_df["sagitta"])

    alpha = sheet.edge_df["arc_chord_angle"]
    beta = sheet.edge_df["chord_orient"]

    # Ok, I admit a fair amount of trial and
    # error to get to the stuff below :-p
    rot = beta - np.sign(alpha) * np.pi / 2
    theta1 = (-alpha + rot) * np.sign(alpha)
    theta2 = (alpha + rot) * np.sign(alpha)

    center_data = pd.DataFrame.from_dict(
        {
            "radius": np.abs(radius),
            "x": center_x,
            "y": center_y,
            "theta1": theta1,
            "theta2": theta2,
        }
    )
    return center_data


def curved_view(sheet, radius_cutoff=1e3):

    center_data = get_arc_data(sheet)
    fig, ax = sheet_view(sheet, **{"edge": {"visible": False}})

    curves = []
    for idx, edge in center_data.iterrows():
        if edge["radius"] > radius_cutoff:
            st = sheet.edge_df.loc[idx, ["srce", "trgt"]]
            xy = sheet.vert_df.loc[st, sheet.coords]
            patch = PathPatch(Path(xy))
        else:
            patch = Arc(
                edge[["x", "y"]],
                2 * edge["radius"],
                2 * edge["radius"],
                theta1=edge["theta1"] * 180 / np.pi,
                theta2=edge["theta2"] * 180 / np.pi,
            )
        curves.append(patch)
    ax.add_collection(PatchCollection(curves, False, **{"facecolors": "none"}))
    ax.autoscale()
    return fig, ax


def plot_junction(eptm, edge_index, coords=["x", "y"]):
    """Plots local graph around a junction, for debugging purposes
    """
    v10, v11 = eptm.edge_df.loc[edge_index, ["srce", "trgt"]]
    fig, ax = plt.subplots()
    ax.scatter(*eptm.vert_df.loc[[v10, v11], coords].values.T, marker="+", s=300)
    v10_out = set(eptm.edge_df[eptm.edge_df["srce"] == v10]["trgt"]) - {v11}
    v11_out = set(eptm.edge_df[eptm.edge_df["srce"] == v11]["trgt"]) - {v10}
    verts = v10_out.union(v11_out)

    ax.scatter(*eptm.vert_df.loc[v10_out, coords].values.T)
    ax.scatter(*eptm.vert_df.loc[v11_out, coords].values.T)

    for _, edge in eptm.edge_df.query(f"srce == {v10}").iterrows():
        ax.plot(
            edge[["s" + coords[0], "t" + coords[0]]],
            edge[["s" + coords[1], "t" + coords[1]]],
            lw=3,
            alpha=0.3,
            c="r",
        )

    for _, edge in eptm.edge_df.query(f"srce == {v11}").iterrows():
        ax.plot(
            edge[["s" + coords[0], "t" + coords[0]]],
            edge[["s" + coords[1], "t" + coords[1]]],
            "k--",
        )

    for v in verts:
        for _, edge in eptm.edge_df.query(f"srce == {v}").iterrows():
            if edge["trgt"] in {v10, v11}:
                continue
            ax.plot(
                edge[["s" + coords[0], "t" + coords[0]]],
                edge[["s" + coords[1], "t" + coords[1]]],
                "k",
                lw=0.4,
            )

    fig.set_size_inches(12, 12)
    return fig, ax



"""
Mesoderm invagination event module
=======================


"""

import random
import numpy as np

from ...utils.decorators import face_lookup
from .actions import relax, contract, ab_pull
from .basic_events import contraction


default_constriction_spec = {
    "face_id": -1,
    "face": -1,
    "contract_rate": 2,
    "critical_area": 1e-2,
    "radial_tension": 1.0,
    "contract_neighbors": True,
    "critical_area_neighbors": 10,
    "contract_span": 2,
    "basal_contract_rate": 1.001,
    "current_traction": 0,
    "max_traction": 30,
    "contraction_column": "contractility",
}


@face_lookup
def constriction(sheet, manager, **kwargs):
    """Constriction process
    This function corresponds to the process called "apical constriction"
    in the manuscript
    The cell undergoing delamination first contracts its apical
    area until it reaches a critical area. A probability
    dependent to the apical area allow an apico-basal
    traction of the cell. The cell can pull during max_traction
    time step, not necessarily consecutively.
    Parameters
    ----------
    sheet : a :class:`tyssue.sheet` object
    manager : a :class:`tyssue.events.EventManager` object
    face_id : int
       the Id of the face undergoing delamination.
    contract_rate : float, default 2
       rate of increase of the face contractility.
    critical_area : float, default 1e-2
       face's area under which the cell starts loosing sides.
    radial_tension : float, default 1.
       tension applied on the face vertices along the
       apical-basal axis.
    contract_neighbors : bool, default `False`
       if True, the face contraction triggers contraction of the neighbor
       faces.
    contract_span : int, default 2
       rank of neighbors contracting if contract_neighbor is True. Contraction
       rate for the neighbors is equal to `contract_rate` devided by
       the rank.
    """
    constriction_spec = default_constriction_spec
    constriction_spec.update(**kwargs)

    # initialiser une variable face
    # aller chercher la valeur dans le dictionnaire à chaque fois ?
    face = constriction_spec["face"]
    contract_rate = constriction_spec["contract_rate"]
    current_traction = constriction_spec["current_traction"]

    if "is_relaxation" in sheet.face_df.columns:
        if sheet.face_df.loc[face, "is_relaxation"]:
            relax(sheet, face, contract_rate, constriction_spec["contraction_column"])

    if sheet.face_df.loc[face, "is_mesoderm"]:
        face_area = sheet.face_df.loc[face, "area"]

        if face_area > constriction_spec["critical_area"]:
            contract(
                sheet,
                face,
                contract_rate,
                True,
                constriction_spec["contraction_column"],
            )
            # if sheet.face_df.loc[face, 'prefered_area'] > 6:
            #    sheet.face_df.loc[face, 'prefered_area'] -= 0.5
            # increase_linear_tension(sheet, face, contract_rate)

            if (constriction_spec["contract_neighbors"]) & (
                face_area < constriction_spec["critical_area_neighbors"]
            ):
                neighbors = sheet.get_neighborhood(
                    face, constriction_spec["contract_span"]
                ).dropna()
                neighbors["id"] = sheet.face_df.loc[neighbors.face, "id"].values

                # remove cell which are not mesoderm
                ectodermal_cell = sheet.face_df.loc[neighbors.face][
                    ~sheet.face_df.loc[neighbors.face, "is_mesoderm"]
                ].id.values

                neighbors = neighbors.drop(
                    neighbors[neighbors.id.isin(ectodermal_cell)].index
                )

                manager.extend(
                    [
                        (
                            contraction,
                            _neighbor_contractile_increase(neighbor, constriction_spec),
                        )  # TODO: check this
                        for _, neighbor in neighbors.iterrows()
                    ]
                )

        proba_tension = np.exp(-face_area / constriction_spec["critical_area"])
        aleatory_number = random.uniform(0, 1)
        if current_traction < constriction_spec["max_traction"]:
            if aleatory_number < proba_tension:
                current_traction = current_traction + 1
                ab_pull(sheet, face, constriction_spec["radial_tension"], True)
                constriction_spec.update({"current_traction": current_traction})

    manager.append(constriction, **constriction_spec)


def _neighbor_contractile_increase(neighbor, constriction_spec):

    contract = constriction_spec["contract_rate"]
    basal_contract = constriction_spec["basal_contract_rate"]

    increase = (
        -(contract - basal_contract) / constriction_spec["contract_span"]
    ) * neighbor["order"] + contract

    specs = {
        "face_id": neighbor["id"],
        "contractile_increase": increase,
        "critical_area": constriction_spec["critical_area"],
        "max_contractility": 50,
        "contraction_column": constriction_spec["contraction_column"],
        "multiple": True,
        "unique": False,
    }

    return specs



from ..geometry import planar_periodic_sheet


def planar_periodic8x8():

    spec = planar_periodic_sheet()
    spec["settings"].update({"boundaries": {"x": [-0.1, 8], "y": [-0.1, 8]}})
    return spec



"""
Small event module
=======================


"""
import numpy as np
import logging

logger = logging.getLogger(__name__)


from ...utils.decorators import face_lookup
from ...geometry.sheet_geometry import SheetGeometry
from ...topology.sheet_topology import cell_division

from .actions import (
    grow,
    contract,
    exchange,
    remove,
    merge_vertices,
    detach_vertices,
    increase,
    decrease,
    increase_linear_tension,
)


def reconnect(sheet, manager, **kwargs):
    """Performs reconnections (vertex merging / splitting) following Finegan et al. 2019

    kwargs overwrite their corresponding `sheet.settings` entries

    Keyword Arguments
    -----------------
    threshold_length : the threshold length at which vertex merging is performed
    p_4 : the probability per unit time to perform a detachement from a rank 4 vertex
    p_5p : the probability per unit time to perform a detachement from a rank 5 or more vertex


    See Also
    --------

    **The tricellular vertex-specific adhesion molecule Sidekick
    facilitates polarised cell intercalation during Drosophila axis
    extension** _Tara M Finegan, Nathan Hervieux, Alexander
    Nestor-Bergmann, Alexander G. Fletcher, Guy B Blanchard, Benedicte
    Sanson_ bioRxiv 704932; doi: https://doi.org/10.1101/704932

    """
    sheet.settings.update(kwargs)
    nv = sheet.Nv
    merge_vertices(sheet)
    if nv != sheet.Nv:
        logger.info(f"Merged {nv - sheet.Nv+1} vertices")
    nv = sheet.Nv
    try:
        detach_vertices(sheet)
    except ValueError:
        logger.info(f"Failed to detach, skipping")
        pass
    if nv != sheet.Nv:
        logger.info(f"Detached {sheet.Nv - nv} vertices")

    manager.append(reconnect, **kwargs)


default_division_spec = {
    "face_id": -1,
    "face": -1,
    "growth_rate": 0.1,
    "critical_vol": 2.0,
    "geom": SheetGeometry,
}


@face_lookup
def division(sheet, manager, **kwargs):
    """Cell division happens through cell growth up to a critical volume,
    followed by actual division of the face.

    Parameters
    ----------
    sheet : a `Sheet` object
    manager : an `EventManager` instance
    face_id : int,
      index of the mother face
    growth_rate : float, default 0.1
      rate of increase of the prefered volume
    critical_vol : float, default 2.
      volume at which the cells stops to grow and devides

    """
    division_spec = default_division_spec
    division_spec.update(**kwargs)

    face = division_spec["face"]

    division_spec["critical_vol"] *= sheet.specs["face"]["prefered_vol"]

    print(sheet.face_df.loc[face, "vol"], division_spec["critical_vol"])

    if sheet.face_df.loc[face, "vol"] < division_spec["critical_vol"]:
        grow(sheet, face, division_spec["growth_rate"])
        manager.append(division, **division_spec)
    else:
        daughter = cell_division(sheet, face, division_spec["geom"])
        sheet.face_df.loc[daughter, "id"] = sheet.face_df.id.max() + 1


default_contraction_spec = {
    "face_id": -1,
    "face": -1,
    "contractile_increase": 1.0,
    "critical_area": 1e-2,
    "max_contractility": 10,
    "multiply": False,
    "contraction_column": "contractility",
    "unique": True,
}


@face_lookup
def contraction(sheet, manager, **kwargs):
    """Single step contraction event
    """
    contraction_spec = default_contraction_spec
    contraction_spec.update(**kwargs)
    face = contraction_spec["face"]

    if (sheet.face_df.loc[face, "area"] < contraction_spec["critical_area"]) or (
        sheet.face_df.loc[face, contraction_spec["contraction_column"]]
        > contraction_spec["max_contractility"]
    ):
        return
    increase(
        sheet,
        "face",
        face,
        contraction_spec["contractile_increase"],
        contraction_spec["contraction_column"],
        contraction_spec["multiply"],
    )


default_type1_transition_spec = {
    "face_id": -1,
    "face": -1,
    "critical_length": 0.1,
    "geom": SheetGeometry,
}


@face_lookup
def type1_transition(sheet, manager, **kwargs):
    """Custom type 1 transition event that tests if
    the the shorter edge of the face is smaller than
    the critical length.
    """
    type1_transition_spec = default_type1_transition_spec
    type1_transition_spec.update(**kwargs)
    face = type1_transition_spec["face"]

    edges = sheet.edge_df[sheet.edge_df["face"] == face]
    if min(edges["length"]) < type1_transition_spec["critical_length"]:
        exchange(sheet, face, type1_transition_spec["geom"])


default_face_elimination_spec = {"face_id": -1, "face": -1, "geom": SheetGeometry}


@face_lookup
def face_elimination(sheet, manager, **kwargs):
    """Removes the face with if face_id from the sheet
    """
    face_elimination_spec = default_face_elimination_spec
    face_elimination_spec.update(**kwargs)
    remove(sheet, face_elimination_spec["face"], face_elimination_spec["geom"])


default_check_tri_face_spec = {"geom": SheetGeometry}


def check_tri_faces(sheet, manager, **kwargs):
    """Three neighbourghs cell elimination
    Add all cells with three neighbourghs in the manager
    to be eliminated at the next time step.
    Parameters
    ----------
    sheet : a :class:`tyssue.sheet` object
    manager : a :class:`tyssue.events.EventManager` object
    """
    check_tri_faces_spec = default_check_tri_face_spec
    check_tri_faces_spec.update(**kwargs)

    tri_faces = sheet.face_df[(sheet.face_df["num_sides"] < 4)].id
    manager.extend(
        [
            (face_elimination, {"face_id": f, "geom": check_tri_faces_spec["geom"]})
            for f in tri_faces
        ]
    )


default_contraction_line_tension_spec = {
    "face_id": -1,
    "face": -1,
    "shrink_rate": 1.05,
    "contractile_increase": 1.0,
    "critical_area": 1e-2,
    "max_contractility": 10,
    "multiply": True,
    "contraction_column": "line_tension",
    "unique": True,
}


@face_lookup
def contraction_line_tension(sheet, manager, **kwargs):
    """
    Single step contraction event
    """
    contraction_spec = default_contraction_line_tension_spec
    contraction_spec.update(**kwargs)
    face = contraction_spec["face"]

    if sheet.face_df.loc[face, "area"] < contraction_spec["critical_area"]:
        return

    # reduce prefered_area
    decrease(
        sheet,
        "face",
        face,
        contraction_spec["shrink_rate"],
        col="prefered_area",
        divide=True,
        bound=contraction_spec["critical_area"] / 2,
    )

    increase_linear_tension(
        sheet,
        face,
        contraction_spec["contractile_increase"],
        multiply=contraction_spec["multiply"],
        isotropic=True,
        limit=100,
    )



"""
Basic monolayer event module
=======================


"""

import logging

logger = logging.getLogger(__name__)

from ..sheet.actions import merge_vertices, detach_vertices


def grow(monolayer, cell, grow_rate):
    """Multiplies the equilibrium volume of face
    by a factor (1+shrink_rate)
    """
    factor = 1 + grow_rate
    monolayer.cell_df.loc[cell, "prefered_vol"] *= factor
    monolayer.cell_df.loc[cell, "prefered_area"] *= factor ** (2 / 3)


def shrink(monolayer, cell, shrink_rate):
    """Divides the equilibrium volume of the cell
    by a factor (1+shrink_rate) and its equilibrium area
    by (1+shrink_rate)^2/3
    """
    factor = 1 + shrink_rate
    monolayer.cell_df.loc[cell, "prefered_vol"] /= factor
    monolayer.cell_df.loc[cell, "prefered_area"] /= factor ** (2 / 3)


def contract(
    monolayer,
    face,
    contractile_increase,
    multiply=False,
    contraction_column="contractility",
):
    """
    Contract the face by increasing the 'contractility' parameter
    by contractile_increase
    """
    if multiply:
        monolayer.face_df.loc[face, contraction_column] *= contractile_increase
    else:
        monolayer.face_df.loc[face, contraction_column] += contractile_increase


def relax(monolayer, face, contractile_decrease, contraction_column="contractility"):
    initial_contractility = 1.12
    new_contractility = (
        monolayer.face_df.loc[face, contraction_column] / contractile_decrease
    )

    if new_contractility >= (initial_contractility / 2):
        monolayer.face_df.loc[face, contraction_column] = new_contractility
        monolayer.face_df.loc[face, "prefered_area"] *= contractile_decrease


def contract_apical_face(
    monolayer,
    face_id,
    contractile_increase=1.0,
    critical_area=1e-2,
    max_contractility=50,
    multiply=False,
    contraction_column="contractility",
):
    """Single step contraction event for apical face only
    """
    face = monolayer.idx_lookup(face_id, "face")
    if face is None:
        return
    if (
        (monolayer.face_df.loc[face, "segment"] != "apical")
        or (monolayer.face_df.loc[face, "area"] < critical_area)
        or (monolayer.face_df.loc[face, contraction_column] > max_contractility)
    ):
        return
    contract(monolayer, face, contractile_increase, multiply, contraction_column)


def ab_pull(monolayer, cell, radial_tension, distributed=False):
    """Adds a linear tension to the apical-to-basal edges
    of a cell
    """
    cell_edges = monolayer.edge_df[monolayer.edge_df["cell"] == cell]
    lateral_edges = cell_edges[cell_edges["segment"] == "lateral"]
    srce_segment = monolayer.upcast_srce(monolayer.vert_df["segment"]).loc[
        lateral_edges.index
    ]
    trgt_segment = monolayer.upcast_trgt(monolayer.vert_df["segment"]).loc[
        lateral_edges.index
    ]

    ab_edges = lateral_edges[
        (srce_segment == "apical") & (trgt_segment == "basal")
    ].index
    ba_edges = lateral_edges[
        (trgt_segment == "apical") & (srce_segment == "basal")
    ].index

    if distributed:
        new_tension = radial_tension / (len(ab_edges) + len(ba_edges))
    else:
        new_tension = radial_tension
    monolayer.edge_df.loc[ab_edges, "line_tension"] += new_tension
    monolayer.edge_df.loc[ba_edges, "line_tension"] += new_tension


def ab_pull_edge(monolayer, cell_edges, radial_tension, distributed=False):
    """Adds a linear tension to the apical-to-basal edges
    of a cell
    """

    if distributed:
        new_tension = radial_tension / (len(cell_edges))
    else:
        new_tension = radial_tension
    monolayer.edge_df.loc[cell_edges, "line_tension"] += new_tension



"""Viscosity based ODE solver


"""
import logging
import numpy as np
import pandas as pd
import warnings

from itertools import count
from scipy.integrate import solve_ivp


from ..core.history import History
from ..behaviors.event_manager import EventManager
from ..behaviors.sheet.basic_events import reconnect


log = logging.getLogger(__name__)
MAX_ITER = 1000


def set_pos(eptm, geom, pos):
    """Updates the vertex position of the :class:`Epithelium` object.

    Assumes that pos is passed as a 1D array to be reshaped as (eptm.Nv, eptm.dim)

    """
    log.debug("set pos")
    eptm.vert_df.loc[eptm.active_verts, eptm.coords] = pos.reshape((-1, eptm.dim))
    geom.update_all(eptm)


class EulerSolver:
    """Explicit Euler solver



    """

    def __init__(
        self,
        eptm,
        geom,
        model,
        history=None,
        auto_reconnect=False,
        manager=None,
        bounds=None,
        with_t1=False,
        with_t3=False,
    ):
        """creates an instance of EulerSolver

        Parameters
        ----------
        eptm : a :class:`tyssue.Epithelium` instance
        geom : a Geometry class
        model : a Model class
        history : a :class:`tyssue.History` or :class:`tyssue.Hdf5History` instance
        auto_reconnect : bool
            if True, will automatically perform reconnections, default False
        manager : a :class:`tyssue.EventManager` instance
        bounds : tuple of (min, max),
            bonds the displacement of the vertices at each time step

        """

        self._set_pos = set_pos
        if with_t1:
            warnings.warn("with_t1 is deprecated and has no effect")
            # self._set_pos = auto_t1(self._set_pos)
        if with_t3:
            warnings.warn("with_t3 is deprecated and has no effect")
            # self._set_pos = auto_t3(self._set_pos)

        # self.rearange = with_t1 or with_t3
        # self.with_t3 = with_t3
        self.eptm = eptm
        self.geom = geom
        self.model = model
        if history is None:
            self.history = History(eptm)
        else:
            self.history = history
        self.prev_t = 0

        if auto_reconnect:
            if manager is None:
                manager = EventManager()
            if not "reconnect" in [n[0].__name__ for n in manager.next]:
                manager.append(reconnect)

        self.manager = manager
        self.bounds = bounds

    @property
    def current_pos(self):
        return self.eptm.vert_df[self.eptm.coords].values.ravel()

    def set_pos(self, pos):
        """Updates the eptm vertices position
        """
        return self._set_pos(self.eptm, self.geom, pos)

    def record(self, t):
        self.history.record(time_stamp=t)

    def solve(self, tf, dt, on_topo_change=None, topo_change_args=()):
        """Solves the system of differential equations from the current time
        to tf with steps of dt with a forward Euler method.

        Parameters
        ----------
        tf : float, final time when we stop solving
        dt : float, time step
        on_topo_change : function, optional, default None
             function of `self.eptm`
        topo_change_args : tuple, arguments passed to `on_topo_change`

        """
        for t in np.arange(self.prev_t, tf + dt, dt):
            pos = self.current_pos
            dot_r = self.ode_func(t, pos)
            if self.bounds is not None:
                dot_r = np.clip(dot_r, *self.bounds)
            pos = pos + dot_r * dt
            self.set_pos(pos)
            self.prev_t = t
            if self.manager is not None:
                self.manager.execute(self.eptm)
                self.geom.update_all(self.eptm)
                self.manager.update()

            if self.eptm.topo_changed:
                log.info("Topology changed")
                if on_topo_change is not None:
                    on_topo_change(*topo_change_args)
                self.eptm.topo_changed = False
            self.record(t)

    def ode_func(self, t, pos):
        """Computes the models' gradient.


        Returns
        -------
        dot_r : 1D np.ndarray of shape (self.eptm.Nv * self.eptm.dim, )

        .. math::
        \frac{dr_i}{dt} = \frac{\nabla U_i}{\eta_i}

        """

        grad_U = -self.model.compute_gradient(self.eptm)
        return (grad_U.values / self.eptm.vert_df["viscosity"].values[:, None]).ravel()


class IVPSolver:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError(
            """It is not yet clear how to use scipy's `solve_ivp` with topology changes

Previous attempts where made but turned out to be clumsy..."""
        )



from ..utils.utils import to_nd


def length_grad(sheet):
    """returns -(dx/l, dy/l, dz/l), ie grad_i(l_ij))
    """
    grad_lij = -(
        sheet.edge_df[sheet.dcoords] / to_nd(sheet.edge_df["length"], sheet.dim)
    )
    grad_lij.columns = sheet.coords
    return grad_lij



import os
import warnings
import traceback
import pandas as pd
import numpy as np

from pathlib import Path

from collections import defaultdict
from .sheet import Sheet
from .objects import Epithelium


def _filter_columns(cols_hist, cols_in, element):
    if not set(cols_hist).issubset(cols_in):
        warnings.warn(
            f"""
Columns {set(cols_hist).difference(cols_in)} are in the history
 {element} dataframe but not in the sheet {element} dataframe.

These non existent columns will not be saved."""
        )
        cols_hist = set(cols_hist).intersection(cols_in)
    return list(cols_hist)


class History:
    """ This class handles recording and retrieving time series
    of sheet objects.

    """

    def __init__(self, sheet, save_every=None, dt=None, extra_cols=None):
        """Creates a `SheetHistory` instance.

        Parameters
        ----------
        sheet : a :class:`Sheet` object which we want to record
        save_every : float, set every time interval to save the sheet
        dt : float, time step
        extra_cols : dictionnary with sheet.datasets as keys and list of
            columns as values. Default None
        """
        if extra_cols is None:
            extra_cols = defaultdict(list)
        else:
            extra_cols = defaultdict(list, **extra_cols)

        self.sheet = sheet

        self.time = 0.0
        self.index = 0
        if save_every is not None:
            self.save_every = save_every
            self.dt = dt
        else:
            self.save_every = None

        self.datasets = {}
        self.columns = {}
        vcols = sheet.coords + extra_cols["vert"]
        vcols = list(set(vcols))
        self.vcols = _filter_columns(vcols, sheet.vert_df.columns, "vertex")
        _vert_h = sheet.vert_df[self.vcols].reset_index(drop=False)
        if not "time" in self.vcols:
            _vert_h["time"] = 0
        self.datasets["vert"] = _vert_h
        self.columns["vert"] = self.vcols

        fcols = extra_cols["face"]
        self.fcols = _filter_columns(fcols, sheet.face_df.columns, "face")
        _face_h = sheet.face_df[self.fcols].reset_index(drop=False)
        if not "time" in self.fcols:
            _face_h["time"] = 0
        self.datasets["face"] = _face_h
        self.columns["face"] = self.fcols

        if sheet.cell_df is not None:
            ccols = extra_cols["cell"]
            self.ccols = _filter_columns(ccols, sheet.cell_df.columns, "cell")
            _cell_h = sheet.cell_df[self.ccols].reset_index(drop=False)
            if not "time" in self.ccols:
                _cell_h["time"] = 0
            self.datasets["cell"] = _cell_h
            self.columns["cell"] = self.ccols
            extra_cols["edge"].append("cell")

        ecols = ["srce", "trgt", "face"] + extra_cols["edge"]
        ecols = list(set(ecols))
        self.ecols = _filter_columns(ecols, sheet.edge_df.columns, "edge")
        _edge_h = sheet.edge_df[self.ecols].reset_index(drop=False)
        if not "time" in self.ecols:
            _edge_h["time"] = 0
        self.datasets["edge"] = _edge_h
        self.columns["edge"] = self.ecols

    def __len__(self):
        return self.time_stamps.__len__()

    def to_archive(self, hf5file):
        """Saves the history to a HDF file

        This file can later be accessed again with the `HistoryHdf5.from_archive`
        class method

        """
        with pd.HDFStore(hf5file, "a") as store:
            for key, df in self.datasets.items():
                store.append(key=key, value=df, data_columns=["time"])

    @property
    def time_stamps(self):
        return self.datasets["vert"]["time"].unique()

    @property
    def vert_h(self):
        return self.datasets["vert"]

    @property
    def edge_h(self):
        return self.datasets["edge"]

    @property
    def face_h(self):
        return self.datasets["face"]

    @property
    def cell_h(self):
        return self.datasets.get("cell", None)

    def record(self, to_record=None, time_stamp=None):
        """Appends a copy of the sheet datasets to the history instance.

        Parameters
        ----------
        to_report : deprecated
        """
        if to_record is not None:
            warnings.warn("Deprecated all the data will be saved")

        if time_stamp is not None:
            self.time = time_stamp
        else:
            self.time += 1

        if (self.save_every is None) or (
            self.index % (int(self.save_every / self.dt)) == 0
        ):
            for element in self.datasets:
                hist = self.datasets[element]
                cols = self.columns[element]
                df = self.sheet.datasets[element][cols].reset_index(drop=False)
                if not "time" in cols:
                    times = pd.Series(np.ones((df.shape[0],)) * self.time, name="time")
                    df = pd.concat([df, times], ignore_index=False, axis=1, sort=False)
                if self.time in hist["time"]:
                    # erase previously recorded time point
                    hist = hist[hist["time"] != self.time]

                hist = pd.concat([hist, df], ignore_index=True, axis=0, sort=False)

                self.datasets[element] = hist

        self.index += 1

    def retrieve(self, time):
        """Return datasets at time `time`.

        If a specific dataset was not recorded at time time, the closest record before that
        time is used.
        """
        if time > self.datasets["vert"]["time"].values[-1]:
            warnings.warn(
                """
The time argument you requested is bigger than the maximum recorded time,
are you sure you pass time in parameter and not an index ?
"""
            )
        sheet_datasets = {}
        for element in self.datasets:
            hist = self.datasets[element]
            cols = self.columns[element]
            df = _retrieve(hist, time)
            df = df.set_index(element)[cols]
            sheet_datasets[element] = df

        return type(self.sheet)(
            f"{self.sheet.identifier}_{time:04.3f}", sheet_datasets, self.sheet.specs
        )

    def __iter__(self):

        for t in self.time_stamps:
            sheet = self.retrieve(t)
            yield t, sheet


class HistoryHdf5(History):
    """ This class handles recording and retrieving time series
    of sheet objects.

    """

    def __init__(
        self,
        sheet=None,
        save_every=None,
        dt=None,
        extra_cols=None,
        hf5file="",
        overwrite=False,
    ):
        """Creates a `SheetHistory` instance.

        Parameters
        ----------
        sheet : a :class:`Sheet` object which we want to record
        save_every : float, set every time interval to save the sheet
        dt : float, time step
        extra_cols : dictionnary with sheet.datasets as keys and list of
            columns as values. Default None
        hf5file : string, define the path of the HDF5 file
        overwrite : bool, Overwrite or not the file if it is already exist. Default False
        """
        if not hf5file:
            warnings.warn(
                "No directory is given. The HDF5 file will be saved in the working directory as out.hf5."
            )
            self.hf5file = Path(os.getcwd()) / "out.hf5"
        else:
            self.hf5file = Path(hf5file)

        if self.hf5file.exists():
            if overwrite:
                tb = traceback.extract_stack(limit=2)
                if not "from_archive" in tb[0].name:
                    warnings.warn(
                        "The file already exist and will be overwritten."
                        " This is normal if you reopened an archive"
                    )
            else:
                expand = 0
                while True:
                    new_hf5file = self.hf5file.parent / self.hf5file.name.replace(
                        self.hf5file.suffix, f"{expand}{self.hf5file.suffix}"
                    )
                    expand += 1
                    if new_hf5file.exists():
                        continue
                    else:
                        self.hf5file = new_hf5file
                        warnings.warn(
                            "The file already exist and the new filename is {}".format(
                                new_hf5file
                            )
                        )
                        break
        if sheet is None:
            last = self.time_stamps[-1]
            with pd.HDFStore(self.hf5file, "r") as file:
                keys = file.keys()
            if "\cell" in keys:
                sheet = Epithelium

        History.__init__(self, sheet, save_every, dt, extra_cols)
        self.dtypes = {
            k: df[self.columns[k]].dtypes for k, df in sheet.datasets.items()
        }

    @classmethod
    def from_archive(cls, hf5file, columns=None, eptm_class=None):
        datasets = {}
        settings = {}
        hf5file = Path(hf5file)
        with pd.HDFStore(hf5file, "r") as store:
            keys = [k.strip("/") for k in store.keys()]
            if columns is None:
                # read everything
                columns = {k: None for k in keys}

            if eptm_class is None:
                eptm_class = Epithelium if "cell" in keys else Sheet

            last = store.select("vert", columns=["time"]).iloc[-1]["time"]
            for key in keys:
                if key == "settings":
                    settings = store[key]
                    continue
                df = store.select(key, where=f"time == {last}", columns=columns[key])
                datasets[key] = df

        eptm = eptm_class(hf5file.name, datasets)
        eptm.settings.update(settings)
        return cls(sheet=eptm, hf5file=hf5file, overwrite=True)

    @property
    def time_stamps(self):
        with pd.HDFStore(self.hf5file, "r") as file:
            times = file.select("vert", columns=["time"])["time"].unique()
        return times

    def record(self, to_record=None, time_stamp=None, sheet=None):
        """Appends a copy of the sheet datasets to the history HDF file.

        Parameters
        ----------
        to_report : Deprecated - list of strings
            the datasets from self.sheet to be saved
        sheet: a :class:`Sheet` object which we want to record. This argument can
        be used if the sheet object is different at each time point.

        """
        if to_record is not None:
            warnings.warn("Deprecated, all the datasets will be saved anyway")

        if sheet is not None:
            self.sheet = sheet

        if time_stamp is not None:
            self.time = time_stamp
        else:
            self.time += 1.0

        dtypes_ = {k: df.dtypes for k, df in self.sheet.datasets.items()}

        for element, df in self.sheet.datasets.items():
            diff_col = set(dtypes_[element].keys()).difference(
                set(self.dtypes[element].keys())
            )
            if diff_col:
                warnings.warn(
                    "New columns {} will not be saved in the {} table".format(
                        diff_col, element
                    )
                )
            else:
                if dtypes_[element].to_dict() != self.dtypes[element].to_dict():
                    raise ValueError(
                        "There is a change of datatype in {} table".format(element)
                    )

        if (self.save_every is None) or (
            self.index % (int(self.save_every / self.dt)) == 0
        ):
            for element, df in self.sheet.datasets.items():
                times = pd.Series(np.ones((df.shape[0],)) * self.time, name="time")
                df = df[self.columns[element]]
                df = pd.concat([df, times], ignore_index=False, axis=1, sort=False)

                with pd.HDFStore(self.hf5file, "a") as file:
                    file.append(key=element, value=df, data_columns=["time"])

        self.index += 1

    def retrieve(self, time):
        """Return datasets at time `time`.

        If a specific dataset was not recorded at time time, the closest record before that
        time is used.
        """
        times = self.time_stamps
        if time > times[-1]:
            warnings.warn(
                "The time argument you passed is bigger than the maximum recorded time, are you sure you pass time in parameter and not an index ? "
            )

        time = times[np.argmin(np.abs(times - time))]
        with pd.HDFStore(self.hf5file, "r") as store:
            sheet_datasets = {}
            for element in self.datasets:
                sheet_datasets[element] = store.select(element, where=f"time == {time}")

        return type(self.sheet)(
            f"{self.sheet.identifier}_{time:04.3f}", sheet_datasets, self.sheet.specs
        )


def _retrieve(dset, time):
    times = dset["time"].values
    t = times[np.argmin(np.abs(times - time))]
    return dset[dset["time"] == t]



import numpy as np


def effector_tester(eptm, effector):

    for u in eptm.ucoords:
        eptm.edge_df[u] = 0

    for elem, spec in effector.specs.items():

        for col in spec:
            if col not in eptm.datasets[elem].columns:
                eptm.datasets[elem][col] = 1.0

    energy = effector.energy(eptm)
    assert energy.shape == (eptm.datasets[effector.element].shape[0],)
    assert np.all(np.isfinite(energy.values))

    grad_s, grad_t = effector.gradient(eptm)
    assert np.all(np.isfinite(grad_s))

    if grad_s.shape == (eptm.Nv, eptm.dim):
        assert grad_t is None
    elif grad_s.shape == (eptm.Ne, eptm.dim):
        assert grad_t.shape[0] == eptm.Ne
        assert np.all(np.isfinite(grad_t))
    else:
        raise ValueError(
            f"""
            The computed gradients for effector {effector.label}
            should have shape {(eptm.Ne, eptm.dim)} or {(eptm.Nv, eptm.dim)},
            found {grad_s.shape}.
            """
        )


def model_tester(eptm, model):

    for u in eptm.ucoords:
        eptm.edge_df[u] = 0

    nondim_specs = {elem: {k: 1 for k in spec} for elem, spec in model.specs.items()}
    dim_specs = model.dimensionalize(nondim_specs)
    for elem, spec in model.specs.items():
        assert set(dim_specs[elem].keys()) == set(spec.keys())

    nondim_specs = model.dimensionalize(dim_specs)
    for elem, spec in model.specs.items():
        assert set(nondim_specs[elem].keys()) == set(spec.keys())

    eptm.update_specs(dim_specs, reset=False)
    energies = model.compute_energy(eptm, full_output=True)
    assert len(energies) == len(model.labels)
    energy = model.compute_energy(eptm, full_output=False)
    assert np.isfinite(energy)

    gradients = model.compute_gradient(eptm, components=True)
    assert len(gradients) == len(model.labels)
    gradient = model.compute_gradient(eptm, components=False)
    assert gradient.shape == eptm.vert_df[eptm.coords].shape
    assert np.all(np.isfinite(gradient))



import warnings
import numpy as np

from .sheet_geometry import SheetGeometry

from .utils import rotation_matrix
from ..utils import _to_3d
from ..core.sheet import Sheet


class BulkGeometry(SheetGeometry):
    """Geometry functions for 3D cell arangements
    """

    @classmethod
    def update_all(cls, eptm):
        """
        Updates the eptm geometry by updating:
        * the edge vector coordinates
        * the edge lengths
        * the face centroids
        * the normals to each edge associated face
        * the face areas
        * the cell areas
        * the vertices heights (depends on geometry)
        * the face volumes (depends on geometry)

        """
        cls.update_dcoords(eptm)
        cls.update_length(eptm)
        cls.update_perimeters(eptm)
        cls.update_centroid(eptm)
        cls.update_normals(eptm)
        cls.update_vol(eptm)
        cls.update_areas(eptm)

    @staticmethod
    def update_dcoords(eptm):
        SheetGeometry.update_dcoords(eptm)

    @staticmethod
    def update_vol(eptm):
        """

        """
        face_pos = eptm.edge_df[["f" + c for c in eptm.coords]].values
        cell_pos = eptm.edge_df[["c" + c for c in eptm.coords]].values

        eptm.edge_df["sub_vol"] = (
            np.sum((face_pos - cell_pos) * eptm.edge_df[eptm.ncoords].values, axis=1)
            / 6
        )
        eptm.cell_df["vol"] = eptm.sum_cell(eptm.edge_df["sub_vol"])

    @staticmethod
    def update_areas(eptm):

        # SheetGeometry.update_areas(eptm)
        eptm.edge_df["sub_area"] = (
            np.linalg.norm(eptm.edge_df[eptm.ncoords], axis=1) / 2
        )
        eptm.face_df["area"] = eptm.sum_face(eptm.edge_df["sub_area"])
        eptm.cell_df["area"] = eptm.sum_cell(eptm.edge_df["sub_area"])

    @staticmethod
    def update_centroid(eptm):
        scoords = ["s" + c for c in eptm.coords]
        eptm.face_df[eptm.coords] = eptm.edge_df.groupby("face")[scoords].mean()
        face_pos = eptm.upcast_face(eptm.face_df[eptm.coords]).values
        eptm.edge_df[["f" + c for c in eptm.coords]] = face_pos

        eptm.cell_df[eptm.coords] = eptm.edge_df.groupby("cell")[scoords].mean()
        cell_pos = eptm.upcast_cell(eptm.cell_df[eptm.coords]).values
        eptm.edge_df[["c" + c for c in eptm.coords]] = cell_pos

    @staticmethod
    def validate_face_norms(eptm):

        fcoords = ["f" + c for c in eptm.coords]
        ccoords = ["c" + c for c in eptm.coords]

        face_pos = eptm.edge_df[fcoords].values
        cell_pos = eptm.edge_df[ccoords].values

        r_cf = face_pos - cell_pos
        r_cf["face"] = eptm.edge_df["face"]
        r_cf = r_cf.groupby("face").mean()
        face_norm = eptm.edge_df.groupby("face")[eptm.ncoords].mean()

        proj = (face_norm * r_cf.values).sum(axis=1)
        is_outward = proj >= 0
        return is_outward


class RNRGeometry(BulkGeometry):
    @staticmethod
    def update_centroid(eptm):
        scoords = ["s" + c for c in eptm.coords]
        tcoords = ["t" + c for c in eptm.coords]

        srce_pos = eptm.edge_df[scoords].values
        trgt_pos = eptm.edge_df[tcoords].values
        mid_pos = (srce_pos + trgt_pos) / 2
        weighted_pos = eptm.sum_face(mid_pos * _to_3d(eptm.edge_df["length"]))
        eptm.face_df[eptm.coords] = (
            weighted_pos.values / eptm.face_df["perimeter"].values[:, np.newaxis]
        )

        face_pos = eptm.upcast_face(eptm.face_df[eptm.coords]).values
        eptm.edge_df[["f" + c for c in eptm.coords]] = face_pos

        eptm.cell_df[eptm.coords] = eptm.edge_df.groupby("cell")[scoords].mean()
        cell_pos = eptm.upcast_cell(eptm.cell_df[eptm.coords]).values
        eptm.edge_df[["c" + c for c in eptm.coords]] = cell_pos


class MonolayerGeometry(RNRGeometry):
    @staticmethod
    def basal_apical_axis(eptm, cell):
        """
        Returns a unit vector allong the apical-basal axis of the cell
        """
        edges = eptm.edge_df[eptm.edge_df["cell"] == cell]
        srce_segments = eptm.vert_df.loc[edges["srce"], "segment"]
        srce_segments.index = edges.index
        trgt_segments = eptm.vert_df.loc[edges["trgt"], "segment"]
        trgt_segments.index = edges.index
        ba_edges = edges[(srce_segments == "apical") & (trgt_segments == "basal")]
        return ba_edges[eptm.dcoords].mean()

    @classmethod
    def cell_projected_pos(cls, eptm, cell, psi=0):
        """Returns the positions of the cell vertices
        transformed such that the cell center sits at the
        coordinate system's origin and the basal-apical axis
        is the new `z` axis.
        """
        ab_axis = cls.basal_apical_axis(eptm, cell)
        n_xy = np.linalg.norm(ab_axis[["dx", "dy"]])
        theta = -np.arctan2(n_xy, ab_axis.dz)
        direction = [ab_axis.dy, -ab_axis.dx, 0]
        rot = rotation_matrix(theta, direction)
        cell_verts = set(eptm.edge_df[eptm.edge_df["cell"] == cell]["srce"])
        vert_pos = eptm.vert_df.loc[cell_verts, eptm.coords]
        for c in eptm.coords:
            vert_pos[c] -= eptm.cell_df.loc[cell, c]

        r1 = np.dot(vert_pos, rot)

        if abs(psi) < 1e-6:
            vert_pos[:] = r1
        else:
            vert_pos[:] = np.dot(rotation_matrix(psi, [0, 0, 1]), r1)
        return vert_pos


class MonoLayerGeometry(MonolayerGeometry):
    pass


class ClosedMonolayerGeometry(MonolayerGeometry):
    @classmethod
    def update_all(cls, eptm):
        """
        Updates the eptm geometry by updating:
        * the edge vector coordinates
        * the edge lengths
        * the face centroids
        * the normals to each edge associated face
        * the face areas
        * the cell areas
        * the vertices heights (depends on geometry)
        * the face volumes (depends on geometry)

        """
        super().update_all(eptm)
        cls.update_lumen_vol(eptm)

    @staticmethod
    def update_lumen_vol(eptm):
        """

        """
        lumen_edges = eptm.edge_df[
            eptm.edge_df.segment == eptm.settings.get("lumen_side", "basal")
        ].copy()
        lumen_pos_faces = lumen_edges[["f" + c for c in eptm.coords]].values
        lumen_sub_vol = (
            np.sum((lumen_pos_faces) * lumen_edges[eptm.ncoords].values, axis=1) / 6
        )
        eptm.settings["lumen_vol"] = -lumen_sub_vol.sum()



import numpy as np
import pandas as pd

import vispy as vp
from vispy import app, scene
from ..config.draw import sheet_spec
from ..utils.utils import spec_updater


def face_visual(sheet, coords=None, **draw_specs_kw):

    if coords is None:
        coords = list("xyz")

    draw_specs = sheet_spec()["face"]
    draw_specs.update(draw_specs_kw)
    if "visible" in sheet.face_df.columns:
        vertices = np.concatenate(
            (sheet.face_df[coords], sheet.vert_df[coords]), axis=0
        )

        upcast_visible = sheet.upcast_face(sheet.face_df["visible"])
        visible_edges = sheet.edge_df[upcast_visible]
        # edge indices as (Nf + Nv) * 3 array
        faces = visible_edges[["srce", "trgt", "face"]].to_numpy()
        # The src, trgt, face triangle is correctly oriented
        # both vert_idx cols are shifted by Nf
        faces[:, :2] += sheet.Nf
    else:
        vertices, faces = sheet.triangular_mesh(coords)

    color = None

    if isinstance(draw_specs["color"], str):
        face_colors = None
        color = draw_specs["color"]

    colors = np.asarray(draw_specs["color"])
    if colors.shape == (3,):
        face_colors = pd.DataFrame(index=sheet.edge_df.index, columns=["R", "G", "B"])
        for channel, val in zip("RGB", colors):
            face_colors[channel] = val

    elif colors.shape == (4,):
        face_colors = pd.DataFrame(
            index=sheet.edge_df.index, columns=["R", "G", "B", "A"]
        )
        for channel, val in zip("RGBA", colors):
            face_colors[channel] = val

    elif colors.shape in [(sheet.Nf, 3), (sheet.Nf, 4)]:
        face_colors = pd.DataFrame(
            index=sheet.face_df.index,
            data=colors,
            columns=["R", "G", "B", "A"][: colors.shape[1]],
        )
        face_colors = sheet.upcast_face(face_colors)

    elif colors.shape in [(sheet.Ne, 3), (sheet.Ne, 4)]:
        face_colors = pd.DataFrame(
            index=sheet.edge_df.index,
            data=colors,
            columns=["R", "G", "B", "A"][: colors.shape[1]],
        )

    mesh = scene.visuals.Mesh(
        vertices=vertices, faces=faces, face_colors=face_colors, color=color
    )
    return mesh


def edge_visual(sheet, coords=None, **draw_specs_kw):

    draw_specs = sheet_spec()["edge"]
    draw_specs.update(draw_specs_kw)
    if coords is None:
        coords = list("xyz")

    color = None
    if isinstance(draw_specs["color"], str):
        color = draw_specs["color"]

    else:
        colors = np.asarray(draw_specs["color"])
        if colors.shape == (3,):
            color = pd.DataFrame(
                index=sheet.vert_df.index, columns=["R", "G", "B", "A"]
            )
            for channel, val in zip("RGB", colors):
                color[channel] = val
            color["A"] = draw_specs.get("alpha", 1.0)

        elif colors.shape == (4,):
            color = pd.DataFrame(
                index=sheet.vert_df.index, columns=["R", "G", "B", "A"]
            )
            for channel, val in zip("RGBA", colors):
                color[channel] = val

        elif colors.shape == (sheet.Ne, 3):
            color = pd.DataFrame(
                index=sheet.edge_df.index, data=colors, columns=["R", "G", "B"]
            )
            color["A"] = draw_specs.get("alpha", 1.0)
            # Strangely, color spec is on a vertex, not segment, basis
            color["srce"] = sheet.edge_df["srce"]
            color = color.groupby("srce").mean()

        elif colors.shape == (sheet.Ne, 4):
            color = pd.DataFrame(
                index=sheet.edge_df.index, data=colors, columns=["R", "G", "B", "A"]
            )
            # Strangely, color spec is on a vertex, not segment, basis
            color["srce"] = sheet.edge_df["srce"]
            color = color.groupby("srce").mean()

        elif colors.shape == (sheet.Nv, 3):
            color = pd.DataFrame(
                index=sheet.vert_df.index, data=colors, columns=["R", "G", "B"]
            )
            color["A"] = draw_specs.get("alpha", 1.0)

        elif colors.shape == (sheet.Nv, 4):
            color = pd.DataFrame(
                index=sheet.vert_df.index, data=colors, columns=["R", "G", "B", "A"]
            )
        else:
            raise ValueError(
                """Shape of the color argument doesn't"""
                """ mach the number of edges """
            )

    wire_pos = sheet.vert_df[coords].values
    connect = sheet.edge_df[["srce", "trgt"]].values
    wire = scene.visuals.Line(
        pos=wire_pos, connect=connect, color=color, width=draw_specs["width"]
    )
    return wire


def vp_view(sheet, coords=None, interactive=True, **draw_specs_kw):

    draw_specs = sheet_spec()
    spec_updater(draw_specs, draw_specs_kw)

    if coords is None:
        coords = ["x", "y", "z"]
    canvas = scene.SceneCanvas(keys="interactive", show=True)
    view = canvas.central_widget.add_view()
    view.camera = "turntable"
    view.camera.aspect = 1
    view.bgcolor = vp.color.Color("#222222")

    if draw_specs["face"]["visible"]:
        mesh = face_visual(sheet, coords, **draw_specs["face"])
        view.add(mesh)

    if draw_specs["edge"]["visible"]:
        wire = edge_visual(sheet, coords, **draw_specs["edge"])
        view.add(wire)

    canvas.show()
    view.camera.set_range()
    if interactive:
        app.run()
    return canvas, view



"""
Small event module
=======================


"""
from ...utils.decorators import face_lookup, cell_lookup
from ...geometry.sheet_geometry import SheetGeometry

from .actions import grow, contract
from ..sheet.basic_events import reconnect

default_contraction_spec = {
    "cell_id": -1,
    "cell": -1,
    "side": "all",
    "contractile_increase": 1.0,
    "critical_area": 1e-2,
    "max_contractility": 10,
    "multiple": False,
    "contraction_column": "contractility",
    "unique": True,
}
# Side can be "apical", "basal", "lateral", "all"


@cell_lookup
def contraction(monolayer, manager, **kwargs):
    """
    Single step contraction event
    """
    contraction_spec = default_contraction_spec
    contraction_spec.update(**kwargs)

    cell = contraction_spec["cell"]
    list_face_in_cell = monolayer.get_orbits("cell", "face")

    # Pick face id in function of chosen side
    faces_id = (
        monolayer.face_df[
            (monolayer.face_df.index.isin(list_face_in_cell[cell]))
            & (monolayer.face_df.segment == contraction_spec["side"])
        ]
        .index[0]
        .values
    )

    for f in faces_id:
        if (monolayer.face_df.loc[f, "area"] < contraction_spec["critical_area"]) or (
            monolayer.face_df.loc[f, contraction_spec["contraction_column"]]
            > contraction_spec["max_contractility"]
        ):
            return

        contract(
            monolayer,
            f,
            contraction_spec["contractile_increase"],
            contraction_spec["multiple"],
            contraction_spec["contraction_column"],
        )



"""3D visualisation inside the notebook.
"""
import warnings
import numpy as np
import pandas as pd
from matplotlib import cm
from ipywidgets import interact

from ..config.draw import sheet_spec
from ..utils.utils import spec_updater, get_sub_eptm

try:
    import ipyvolume as ipv
except ImportError:
    print(
        """
This module needs ipyvolume to work.
You can install it with:
$ conda install -c conda-forge ipyvolume
"""
    )


def browse_history(history, coords=["x", "y", "z"], **draw_specs_kw):

    times = history.time_stamps
    num_frames = times.size
    draw_specs = sheet_spec()
    spec_updater(draw_specs, draw_specs_kw)
    sheet = history.retrieve(0)
    ipv.clear()
    fig, meshes = sheet_view(sheet, coords, **draw_specs_kw)

    lim_inf = sheet.vert_df[sheet.coords].min().min()
    lim_sup = sheet.vert_df[sheet.coords].max().max()
    ipv.xyzlim(lim_inf, lim_sup)

    def set_frame(i=0):
        fig.animation = 0
        t = times[i]
        meshes = _get_meshes(history.retrieve(t), coords, draw_specs)
        update_view(fig, meshes)

    ipv.show()
    interact(set_frame, i=(0, num_frames - 1))


def update_view(fig, meshes):

    for old, new in zip(fig.meshes, meshes):
        old.x = new.x
        old.y = new.y
        old.z = new.z
        old.color = new.color
        old.triangles = new.triangles
        old.lines = new.lines


def sheet_view(sheet, coords=["x", "y", "z"], **draw_specs_kw):
    """
    Creates a javascript renderer of the edge lines to be displayed
    in Jupyter Notebooks

    Returns
    -------

    fig: a :class:`ipyvolume.widgets.Figure` widget
    mesh: a :class:`ipyvolume.widgets.Mesh` mesh widget

    """

    # ipv.style.use(["dark", "minimal"])
    draw_specs = sheet_spec()
    spec_updater(draw_specs, draw_specs_kw)
    fig = ipv.gcf()
    fig.meshes = fig.meshes + _get_meshes(sheet, coords, draw_specs)
    box_size = max(*(np.ptp(sheet.vert_df[u]) for u in sheet.coords))
    border = 0.05 * box_size
    lim_inf = sheet.vert_df[sheet.coords].min().min() - border
    lim_sup = sheet.vert_df[sheet.coords].max().max() + border
    ipv.xyzlim(lim_inf, lim_sup)
    return fig, fig.meshes


def view_ipv(sheet, coords=["x", "y", "z"], **edge_specs):
    """
    Creates a javascript renderer of the edge lines to be displayed
    in Jupyter Notebooks

    Returns
    -------

    fig: a :class:`ipyvolume.widgets.Figure` widget
    mesh: a :class:`ipyvolume.widgets.Mesh` mesh widget

    """
    warnings.warn("`view_ipv` is deprecated, use the more generic `sheet_view`")
    mesh = edge_mesh(sheet, coords, **edge_specs)
    fig = ipv.gcf()
    fig.meshes = fig.meshes + [mesh]
    box_size = max(*(np.ptp(sheet.vert_df[u]) for u in sheet.coords))
    border = 0.05 * box_size
    lim_inf = sheet.vert_df[sheet.coords].min().min() - border
    lim_sup = sheet.vert_df[sheet.coords].max().max() + border
    ipv.xyzlim(lim_inf, lim_sup)

    return fig, mesh


def edge_mesh(sheet, coords, **edge_specs):
    """
    Creates a ipyvolume Mesh of the edge lines to be displayed
    in Jupyter Notebooks

    Returns
    -------
    mesh: a :class:`ipyvolume.widgets.Mesh` mesh widget

    """
    spec = sheet_spec()["edge"]
    spec.update(**edge_specs)
    if callable(spec["color"]):
        spec["color"] = spec["color"](sheet)

    if isinstance(spec["color"], str):
        color = spec["color"]
    elif hasattr(spec["color"], "__len__"):
        color = _wire_color_from_sequence(spec, sheet)[:, :3]

    u, v, w = coords
    mesh = ipv.Mesh(
        x=sheet.vert_df[u],
        y=sheet.vert_df[v],
        z=sheet.vert_df[w],
        lines=sheet.edge_df[["srce", "trgt"]].astype(dtype=np.uint32),
        color=color,
    )
    return mesh


def face_mesh(sheet, coords, **face_draw_specs):
    """
    Creates a ipyvolume Mesh of the face polygons
    """
    Ne, Nf = sheet.Ne, sheet.Nf
    if callable(face_draw_specs["color"]):
        face_draw_specs["color"] = face_draw_specs["color"](sheet)

    if isinstance(face_draw_specs["color"], str):
        color = face_draw_specs["color"]

    elif hasattr(face_draw_specs["color"], "__len__"):
        color = _face_color_from_sequence(face_draw_specs, sheet)[:, :3]

    if "visible" in sheet.face_df.columns:
        edges = sheet.edge_df[sheet.upcast_face(sheet.face_df["visible"])].index
        _sheet = get_sub_eptm(sheet, edges)
        if _sheet is not None:
            sheet = _sheet
            if isinstance(color, np.ndarray):
                faces = sheet.face_df["face_o"].values.astype(np.uint32)
                edges = edges.values.astype(np.uint32)
                indexer = np.concatenate([faces, edges + Nf, edges + Ne + Nf])
                color = color.take(indexer, axis=0)

    epsilon = face_draw_specs.get("epsilon", 0)
    up_srce = sheet.edge_df[["s" + c for c in coords]]
    up_trgt = sheet.edge_df[["t" + c for c in coords]]

    Ne, Nf = sheet.Ne, sheet.Nf

    if epsilon > 0:
        up_face = sheet.edge_df[["f" + c for c in coords]].values
        up_srce = (up_srce - up_face) * (1 - epsilon) + up_face
        up_trgt = (up_trgt - up_face) * (1 - epsilon) + up_face

    mesh_ = np.concatenate(
        [sheet.face_df[coords].values, up_srce.values, up_trgt.values]
    )

    triangles = np.vstack(
        [sheet.edge_df["face"], np.arange(Ne) + Nf, np.arange(Ne) + Ne + Nf]
    ).T.astype(dtype=np.uint32)

    mesh = ipv.Mesh(
        x=mesh_[:, 0], y=mesh_[:, 1], z=mesh_[:, 2], triangles=triangles, color=color
    )
    return mesh


def _wire_color_from_sequence(edge_spec, sheet):
    """
    """
    color_ = edge_spec["color"]
    cmap = cm.get_cmap(edge_spec.get("colormap", "viridis"))
    if color_.shape in [(sheet.Nv, 3), (sheet.Nv, 4)]:
        return np.asarray(color_)
    elif color_.shape == (sheet.Nv,):
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((sheet.Nv, 3)) * 0.7
        return cmap((color_ - color_.min()) / np.ptp(color_))

    elif color_.shape in [(sheet.Ne, 3), (sheet.Ne, 4)]:
        color_ = pd.DataFrame(color_, index=sheet.edge_df.index)
        color_["srce"] = sheet.edge_df["srce"]
        color_ = color_.groupby("srce").mean().values
        return color_
    elif color_.shape == (sheet.Ne,):
        color_ = pd.DataFrame(color_, index=sheet.edge_df.index)
        color_["srce"] = sheet.edge_df["srce"]
        color_ = color_.groupby("srce").mean().values.ravel()
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((sheet.Nv, 3)) * 0.7
        return cmap((color_ - color_.min()) / np.ptp(color_))
    else:
        raise ValueError("The 'color' value of the spec doesn't have a correct shape.")


def _face_color_from_sequence(face_spec, sheet):
    color_ = face_spec["color"]

    cmap = cm.get_cmap(face_spec.get("colormap", "viridis"))
    Nf, Ne = sheet.Nf, sheet.Ne
    color_min, color_max = face_spec.get("color_range", (color_.min(), color_.max()))

    face_mesh_shape = Nf + 2 * Ne

    if color_.shape in [(sheet.Nf, 3), (sheet.Nf, 4)]:
        return np.concatenate([color_, color_, color_])

    elif color_.shape == (sheet.Nf,):
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((face_mesh_shape, 3)) * 0.5

        normed = (color_ - color_min) / (color_max - color_min)
        up_color = sheet.upcast_face(normed).values
        return cmap(np.concatenate([normed, up_color, up_color]))

    else:
        raise ValueError(
            "shape of `face_spec['color']` must be either (Nf, 3), (Nf, 4) or (Nf,)"
        )


def _get_meshes(sheet, coords, draw_specs):

    meshes = []
    edge_spec = draw_specs["edge"]
    if edge_spec["visible"]:
        edges = edge_mesh(sheet, coords, **edge_spec)
        meshes.append(edges)
    else:
        edges = None

    face_spec = draw_specs["face"]
    if face_spec["visible"]:
        faces = face_mesh(sheet, coords, **face_spec)
        meshes.append(faces)
    else:
        faces = None
    return meshes



# /bin/env python
"""Quasistatic solver for vertex models

"""

import logging
from itertools import count

from scipy import optimize
from .. import config
from ..collisions import auto_collisions
from ..topology import auto_t1, auto_t3

from .base import TopologyChangeError, set_pos

log = logging.getLogger(__name__)

MAX_ITER = 100


class QSSolver:
    """Quasistatic solver performing a gradient descent on a :class:`tyssue.Epithelium`
    object.

    Methods
    -------
    find_energy_min : energy minimization calling `scipy.optimize.minimize`
    approx_grad : uses `optimize.approx_fprime` to compute an approximated
      gradient.
    check_grad : compares the approximated gradient with the one provided
      by the model
    """

    def __init__(self, with_collisions=False, with_t1=False, with_t3=False):
        """Creates a quasistatic gradient descent solver with optional
        type1, type3 and collision detection and solving routines.

        Parameters
        ----------
        with_collisions : bool, default False
            wheter or not to solve collisions
        with_t1 : bool, default False
            whether or not to solve type 1 transitions at each
            iteration.
        with_t3 : bool, default False
            whether or not to solve type 3 transitions
            (i.e. elimnation of small triangular faces) at each
            iteration.

        Those corrections are applied in this order: first the type 1, then the
        type 3, then the collisions

        """
        self.set_pos = set_pos
        if with_t1:
            self.set_pos = auto_t1(self.set_pos)
        if with_t3:
            self.set_pos = auto_t3(self.set_pos)
        if with_collisions:
            self.set_pos = auto_collisions(self.set_pos)
        self.restart = True
        self.rearange = with_t1 or with_t3
        self.res = {"success": False, "message": "Not Started"}
        self.num_restarts = 0

    def find_energy_min(self, eptm, geom, model, **minimize_kw):
        """Energy minimization function.

        The epithelium's total energy is minimized by displacing its vertices.
        This is a wrapper around `scipy.optimize.minimize`

        Parameters
        ----------
        eptm : a :class:`tyssue.Epithlium` object
        geom : a geometry class
        geom must provide an `update_all` method that takes `eptm`
            as sole argument and updates the relevant geometrical quantities
        model : a model class
            model must provide `compute_energy` and `compute_gradient` methods
            that take `eptm` as first and unique positional argument.

        """
        log.info("initial number of vertices: %i", eptm.Nv)
        settings = config.solvers.quasistatic()
        settings.update(**minimize_kw)

        res = self._minimize(eptm, geom, model, **settings)
        log.info("final number of vertices: %i", eptm.Nv)

        return res

    def _minimize(self, eptm, geom, model, **kwargs):
        for i in count():
            if i == MAX_ITER:
                return self.res
            pos0 = eptm.vert_df.loc[eptm.active_verts, eptm.coords].values.flatten()
            try:
                self.res = optimize.minimize(
                    self._opt_energy,
                    pos0,
                    args=(eptm, geom, model),
                    jac=self._opt_grad,
                    **kwargs
                )
                return self.res
            except TopologyChangeError:
                log.info("TopologyChange")
                self.num_restarts = i + 1

    def _opt_energy(self, pos, eptm, geom, model):
        if self.rearange and eptm.topo_changed:
            # reset switch
            eptm.topo_changed = False
            raise TopologyChangeError("Topology changed before energy evaluation")
        self.set_pos(eptm, geom, pos)
        return model.compute_energy(eptm)

    # The unused arguments bellow are legit, we need the same signature as _opt_energy
    def _opt_grad(self, pos, eptm, geom, model):
        if self.rearange and eptm.topo_changed:
            raise TopologyChangeError("Topology changed before gradient evaluation")
        grad_i = model.compute_gradient(eptm)
        return grad_i.loc[eptm.active_verts].values.ravel()

    def approx_grad(self, eptm, geom, model):
        pos0 = eptm.vert_df.loc[eptm.active_verts, eptm.coords].values.ravel()
        grad = optimize.approx_fprime(pos0, self._opt_energy, 1e-9, eptm, geom, model)
        return grad

    def check_grad(self, eptm, geom, model):
        pos0 = eptm.vert_df.loc[eptm.active_verts, eptm.coords].values.ravel()
        grad_err = optimize.check_grad(
            self._opt_energy, self._opt_grad, pos0.flatten(), eptm, geom, model
        )
        return grad_err



"""
Specific functions for apoptosis vertex model
"""

from ..utils import to_nd
from . import units

from .sheet_gradients import height_grad
from . import effectors
from .factory import model_factory


class ApicoBasalTension(effectors.AbstractEffector):
    """ Effector for the apical-basal tension.

    The energy is proportional to the heigth of the cell
    """

    dimensions = units.line_elasticity
    label = "Apical-basal tension"
    magnitude = "radial_tension"
    element = "vert"
    specs = {"vert": {"radial_tension":0.0, "height": 1.0, "is_active": 1}}

    @staticmethod
    def energy(sheet):
        return sheet.vert_df.eval("height * radial_tension * is_active")

    @staticmethod
    def gradient(sheet):
        grad = to_nd(sheet.vert_df.eval("radial_tension * is_active"), 3) * height_grad(
            sheet
        )
        grad.columns = ["gx", "gy", "gz"]
        return grad, None


SheetApoptosisModel = model_factory(
    [
        effectors.LineTension,
        effectors.FaceContractility,
        ApicoBasalTension,
        effectors.FaceVolumeElasticity,
    ],
    effectors.FaceVolumeElasticity,
)



"""
Core definitions
"""
import warnings
import logging
from collections import deque
from itertools import product
import numpy as np
import pandas as pd
from copy import deepcopy
from ..utils.utils import set_data_columns, spec_updater
from ..utils import connectivity

log = logging.getLogger(name=__name__)


class Epithelium:
    """Base class defining a connective tissue in 2D or 3D.
    """

    def __init__(self, identifier, datasets, specs=None, coords=None, maxbackup=5):
        """Creates an epithelium

        Parameters
        ----------
        identifier : string
        datasets : dictionary of dataframes
            The keys correspond to the different geometrical elements
            constituting the epithelium:

            * `vert` contains a dataframe of vertices,
            * `edge` contains a dataframe of *oriented* half-edges between vertices,
            * `face` contains a dataframe of polygonal faces enclosed by half-edges,
            * `cell` contains a dataframe of polyhedral cells delimited by faces,

        specs : nested dictionnary of specifications
            The first key designs the element name: (`vert`, `edge`, `face`, `cell`),
            corresponding to the respective dataframes attribute in the dataset.
            The second level keys design column names of the above dataframes.
            For exemple:
            .. code::
                specs = {
                    "face": {
                        ## Face Geometry
                        "perimeter": 0.,
                        "area": 0.,
                        ## Coordinates
                        "x": 0.,
                        "y": 0.,
                        "z": 0.,
                        ## Topology
                        "num_sides": 6,
                        ## Masks
                        "is_alive": True},
                    "vert": {
                        ## Coordinates
                        "x": 0.,
                        "y": 0.,
                        "z": 0.,
                        ## Masks
                        "is_active": True},
                    "edge": {
                        ## Connections
                        "srce": 0,
                        "trgt": 1,
                        "face": 0,
                        "cell": 0,
                        ## Coordinates
                        "dx": 0.,
                        "dy": 0.,
                        "dz": 0.,
                        "length": 0.,
                        ## Normals
                        "nx": 0.,
                        "ny": 0.,
                        "nz": 1.}
                    "settings":
                        ## Custom values
                        "geometry": "flat"
                    }

        Note
        ----
        For efficiency reasons, we have to maintain a monotonous RangeIndex
        for each dataset. Thus, **the index of an element can change**,
        and should not be used as an identifier.

        """
        # backup container
        # TODO: pass the max backup number as a config argument
        self._backups = deque(maxlen=maxbackup)

        # each of those has a separate dataframe, as well as entries in
        # the specification files
        _frame_types = {"edge", "vert", "face", "cell"}
        self.identifier = identifier
        if not set(datasets).issubset(_frame_types):
            raise ValueError(
                "The `datasets` dictionnary should"
                " contain keys in {}".format(_frame_types)
            )
        self.datasets = datasets
        self.data_names = list(datasets.keys())
        self.element_names = ["srce", "trgt", "face", "cell"][: len(self.data_names)]
        # Infer specs from the first rows of the datasets
        if specs is None:
            specs = {elem: df.iloc[0].to_dict() for elem, df in datasets.items()}
        if "settings" not in specs:
            specs["settings"] = {}

        self.specs = specs
        if coords is None:
            coords = [c for c in "xyz" if c in datasets["vert"].columns]

        # Add upcast coordinates to specs
        self.specs["edge"].update(
            {
                e + c: 0.0
                for e, c in product("dustfc"[: len(self.data_names) + 2], coords)
            }
        )
        self.specs["face"].update({c: 0.0 for c in coords})
        self.update_specs(specs, reset=False)

        self.coords = coords
        # edge's dx, dy, dz
        self.dcoords = ["d" + c for c in self.coords]
        # edge's unit length vector
        self.ucoords = ["u" + c for c in self.coords]

        self.dim = len(self.coords)
        # edge's normals
        if self.dim == 3:
            self.ncoords = ["n" + c for c in self.coords]
            self.update_specs({"edge": {nc: 0.0 for nc in self.ncoords}}, reset=False)
        else:
            self.update_specs({"edge": {"nz": 0.0}}, reset=False)

        self._bad = None
        self.bbox = None
        if "is_active" in self.vert_df.columns:
            self.active_verts = self.vert_df[self.vert_df.is_active == 1].index
        else:
            self.active_verts = self.vert_df.index
        self.set_bbox()

        self.position_buffer = None
        self.topo_changed = False

    @property
    def vert_df(self):
        """The face :class:`pd.DataFrame` containing vertex associated
        data e.g. the position of each vertex.
        """
        return self.datasets["vert"]

    @vert_df.setter
    def vert_df(self, value):
        self.datasets["vert"] = value

    @property
    def face_df(self):
        """The face :class:`pd.DataFrame` containing face associated
        data e.g. the position of their center or their area
        """
        return self.datasets["face"]

    @face_df.setter
    def face_df(self, value):
        self.datasets["face"] = value

    @property
    def edge_df(self):
        """The edge :class:`pd.DataFrame` containing edge associated
        data e.g. their length.This dataframe also contains the whole
        connexion of the epithelium through the `"srce", "trgt", "face", "cell"`
        indices. In 2D, a half-edge is associated with a single (face, srce, trgt)
        positively oriented triangle. In 3D, the (cell, face, srce, trgt)
        positively oriented terahedron is also unique.
        """
        return self.datasets["edge"]

    @edge_df.setter
    def edge_df(self, value):
        self.datasets["edge"] = value

    @property
    def cell_df(self):
        """The cell :class:`pd.DataFrame` containing cell associated
        data e.g. the position of their center or their volume
        """
        return self.datasets.get("cell", None)

    @cell_df.setter
    def cell_df(self, value):
        self.datasets["cell"] = value

    def copy(self, deep_copy=True):
        """
        Returns a copy of the epithelium

        Parameters
        ----------
        deep_copy : bool, default True
            if True, use a copy of the original object's datasets
            to create the new object. If False, datasets are not copied
        """
        if deep_copy:
            datasets = {element: df.copy() for element, df in self.datasets.items()}
            specs = deepcopy(self.specs)

        else:  # pragma: no cover
            log.info("New epithelium object from %s without deep copy", self.identifier)
            datasets = self.datasets
            specs = self.specs

        identifier = self.identifier + "_copy"

        new = type(self)(identifier, datasets, specs=specs, coords=self.coords)

        return new

    def backup(self):
        """Creates a copy of self and keeps a reference to it
        in the self._backups deque.

        """
        log.info("Backing up")
        self._backups.append(self.copy(deep_copy=True))

    def restore(self):
        """Resets the eptithelium data to its last backed up state

        A copy of the current state prior to restoring is kept in the
        `_bad` attribute for inspection.
        Calling this method multiple times (without calling backup) will
        go back in the epithelium backups.
        """

        log.info("Restoring")
        log.info(
            "a copy of the unrestored epithelium is stored in the `_bad` attribute"
        )
        bck = self._backups.pop()
        self._bad = self.copy(deep_copy=True)
        self.datasets = bck.datasets
        self.specs = bck.specs

    @property
    def settings(self):
        """ Accesses the `specs['settings']` dictionnary
        """
        return self.specs["settings"]

    def update_specs(self, new, reset=False):
        """Recursively updates the `self.specs` nested dictionnary,
        and set the new values to the corresponding columns
        in the datasets. If reset is `True`, existing values
        will be overwritten.
        """
        spec_updater(self.specs, new)
        set_data_columns(self.datasets, new, reset)

    def update_num_sides(self):
        """Updates the number of half-edges around the faces.
        The data is registered in the `"num_sides"` column of
        `self.face_df`.
        """
        self.face_df["num_sides"] = self.edge_df.face.value_counts()

    def update_num_faces(self):
        """Updates the number of faces around the cells.
        The data is registered in the `"num_faces"` column of
        `self.cell_df`.
        """
        self.cell_df["num_faces"] = self.edge_df.groupby("cell").apply(
            lambda df: df["face"].unique().size
        )
        self.cell_df["num_ridges"] = self.edge_df.cell.value_counts()

    def update_rank(self):
        st_connect = connectivity.srce_trgt_connectivity(self)
        self.vert_df["rank"] = ((st_connect + st_connect.T) > 0).sum(axis=0)

    def reset_topo(self):
        """Recomputes the number of sides for the faces and the
        number of faces for the cells.
        """
        log.debug("Resetting topology")
        self.update_num_sides()
        if "is_active" in self.vert_df.columns:
            self.active_verts = self.vert_df[self.vert_df.is_active == 1].index
        if "cell" in self.data_names:
            self.update_num_faces()

    @property
    def Nv(self):
        """The number of vertices in the epithelium
        """
        return self.vert_df.shape[0]

    @property
    def Ne(self):
        """The number of edges in the epithelium
        """
        return self.edge_df.shape[0]

    @property
    def Nf(self):
        """The number of faces in the epithelium
        """
        return self.face_df.shape[0]

    @property
    def Nc(self):
        """The number of cells in the epithelium
        """
        if "cell" in self.data_names:
            return self.cell_df.shape[0]
        elif "face" in self.data_names:
            return self.face_df.shape[0]
        return None

    def _upcast(self, idx, df):

        ## Assumes a flat index
        upcast = df.take(idx)
        try:
            upcast.index = self.edge_df.index
        except AttributeError:
            if len(upcast.shape) == 1:
                upcast = pd.Series(upcast, index=self.edge_df.index)
            else:
                upcast = pd.DataFrame(upcast, index=self.edge_df.index)
        return upcast

    def upcast_cols(self, element, columns):
        """Syntactic sugar to upcast from the epithelium datasets.

        Parameters
        ----------
        element: {'srce'|'trgt'|'face'|'cell'}
           corresponding self.edge_df column over which to index
           if element is 'srce' or 'trgt', the upcast data will be
           taken form self.vert_df
        columns: index
           the column(s) to be taken from the input dataset.

        """
        if element in ["srce", "trgt"]:
            dataset = "vert"
        else:
            dataset = element
        return self._upcast(self.edge_df[element], self.datasets[dataset][columns])

    def upcast_srce(self, df):
        """ Reindexes input data to self.edge_df.index
        by repeating the values for each source entry.

        Parameters
        ----------
        df : :class:`pd.DataFrame`, :class:`pd.Series` :class:`np.ndarray` or string
          The data to be upcasted. If array like, should have `self.Nv` elements.
          If a string is passed it should be a column of `self.vert_df`

        Returns
        -------
        upcast_df : :class:`pd.DataFrame` or :class:`pd.Series`
          The value repeated like the values of `self.edge_df["srce"]`
        """
        if isinstance(df, str):
            df = self.vert_df[df]
        return self._upcast(self.edge_df["srce"], df)

    def upcast_trgt(self, df):
        """ Reindexes input data to self.edge_df.index
        by repeating the values for each target entry

        Parameters
        ----------
        df : :class:`pd.DataFrame`, :class:`pd.Series` :class:`np.ndarray` or string
          The data to be upcasted. If array like, should have `self.Nv` elements.
          If a string is passed it should be a column of `self.vert_df`

        Returns
        -------
        upcast_df : :class:`pd.DataFrame` or :class:`pd.Series`
          The value repeated like the values of `self.edge_df["trgt"]`
        """
        if isinstance(df, str):
            df = self.vert_df[df]
        return self._upcast(self.edge_df["trgt"], df)

    def upcast_face(self, df):
        """ Reindexes input data to self.edge_df.index
        by repeating the values for each face entry

        Parameters
        ----------
        df : :class:`pd.DataFrame`, :class:`pd.Series` :class:`np.ndarray` or string
          The data to be upcasted. If array like, should have `self.Nf` elements.
          If a string is passed it should be a column of `self.face_df`

        Returns
        -------
        upcast_df : :class:`pd.DataFrame` or :class:`pd.Series`
          The value repeated like the values of `self.edge_df["face"]`

        """
        if isinstance(df, str):
            df = self.face_df[df]
        return self._upcast(self.edge_df["face"], df)

    def upcast_cell(self, df):
        """ Reindexes input data to self.edge_df.index
        by repeating the values for each cell entry

        Parameters
        ----------
        df : :class:`pd.DataFrame`, :class:`pd.Series` :class:`np.ndarray` or string
          The data to be upcasted. If array like, should have `self.Nc` elements.
          If a string is passed it should be a column of `self.cell_df`

        Returns
        -------
        upcast_df : :class:`pd.DataFrame` or :class:`pd.Series`
          The value repeated like the values of `self.edge_df["cell"]`
        """
        if isinstance(df, str):
            df = self.cell_df[df]
        return self._upcast(self.edge_df["cell"], df)

    def _lvl_sum(self, df, lvl):
        df_ = df
        if isinstance(df, np.ndarray):
            df_ = pd.DataFrame(df, index=self.edge_df.index)
        elif isinstance(df, pd.Series):
            df_ = df.to_frame()
        elif lvl not in df.columns:
            df_ = df.copy()
        df_[lvl] = self.edge_df[lvl]
        return df_.groupby(lvl).sum()

    def sum_srce(self, df):
        """Sums the values of the edge-indexed dataframe `df` grouped by
        the values of `self.edge_df["srce"]`

        Returns
        -------
        summed : :class:`pd.DataFrame` the summed data, indexed by the source vertices.
        """
        return self._lvl_sum(df, "srce")

    def sum_trgt(self, df):
        """Sums the values of the edge-indexed dataframe `df` grouped by
        the values of `self.edge_df["trgt"]`

        Returns
        -------
        summed : :class:`pd.DataFrame` the summed data, indexed by the source vertices.
        """
        return self._lvl_sum(df, "trgt")

    def sum_face(self, df):
        """Sums the values of the edge-indexed dataframe `df` grouped by
        the values of `self.edge_df["face"]`

        Returns
        -------
        summed : :class:`pd.DataFrame` the summed data, indexed by the source vertices.
        """
        return self._lvl_sum(df, "face")

    def sum_cell(self, df):
        """Sums the values of the edge-indexed dataframe `df` grouped by
        the values of `self.edge_df["cell"]`

        Returns
        -------
        summed : :class:`pd.DataFrame` the summed data, indexed by the source vertices.
        """
        return self._lvl_sum(df, "cell")

    def get_orbits(self, center, periph):
        """Returns a dataframe with a `(center, edge)` MultiIndex with `periph`
        elements.

        Parameters
        ----------
        center : str,
            the name of the center element for example 'face', 'srce'
        periph : str,
            the name of the periphery elements, for example 'trgt', 'cell'

        Example
        -------
        >>> cell_verts = sheet.get_orbits('face', 'srce')
        >>> cell_verts.loc[45]
        edge
        218    75
        219    78
        220    76
        221    81
        222    90
        223    87
        Name: srce, dtype: int64

        """
        orbits = self.edge_df.groupby(center).apply(lambda df: df[periph])
        return orbits

    def idx_lookup(self, elem_id, element):
        """returns the current index of the element with the `"id"` column equal to `elem_id`

        Parameters
        ----------
        elem_id : int
          id of the element to retrieve
        element : {"vert"|"edge"|"face"|"cell"}
          the corresponding dataset.
        """
        df = self.datasets[element]["id"]
        idx = df[df == elem_id].index
        if len(idx):
            return idx[0]
        else:
            return None

    def get_neighbors(self, elem_id, elem="cell"):
        """Returns the indexes of the adjacent elements (cells or faces) of
        the element of index `elem_id`.

        Parameters
        ----------
        elem_id : int
            the index of the central element (a face or a cell)
        element : {'cell' | 'face'}, default 'cell'

        Returns
        -------
        neghbors : set
            the cells (or faces) sharing an edge with the central cell (face)
        """

        topo = self.edge_df[["srce", "trgt", elem]]
        edges = self.edge_df[self.edge_df[elem] == elem_id][["srce", "trgt"]]

        neighbors = set(
            topo[
                (topo["srce"].isin(edges["srce"]) & topo["trgt"].isin(edges["trgt"]))
                | (topo["srce"].isin(edges["trgt"]) & topo["trgt"].isin(edges["srce"]))
            ][elem]
        )
        return neighbors - {elem_id}

    def get_neighborhood(self, elem_id, order, elem="cell"):
        """Returns `elem_id` neighborhood up to a degree of `order`

        For example, if `order` is 2, it wil return the adjacent cells (or faces)
        and theses cells neighbors.

        Returns
        -------
        neighbors : pd.DataFrame with two colums, the index
            of the neighboring cell (face), and it's neighboring order

        """

        neighbors = pd.DataFrame.from_dict({elem: [elem_id], "order": [0]})

        for k in range(order + 1):
            for neigh in neighbors[neighbors["order"] == k - 1][elem]:
                new_neighs = self.get_neighbors(neigh)
                new_neighs = set(new_neighs).difference(neighbors[elem])
                orders = np.ones(len(new_neighs), dtype=np.int) * (k)
                new_neighs = pd.DataFrame.from_dict(
                    {elem: list(new_neighs), "order": orders}, dtype=np.int
                )
                neighbors = pd.concat([neighbors, new_neighs])
        return neighbors.reset_index(drop=True).loc[1:]

    def face_polygons(self, coords=None):
        """ Returns a pd.Series of arrays with the coordinates the face polygons

        Each element of the Series is a (num_sides, num_dims) array of points
        ordered counterclockwise.
        """
        if coords is None:
            coords = self.coords

        def _get_verts_pos(face):
            edges = _ordered_edges(face)
            return np.array([self.vert_df.loc[idx[0], coords] for idx in edges])

        polys = self.edge_df.groupby("face").apply(_get_verts_pos).dropna()
        return polys

    def validate(self):
        """returns True if the mesh is validated

        e.g. has only closed polygons and polyhedra
        """
        return np.alltrue(self.get_valid())

    def get_valid(self):
        """Set the 'is_valid' column to true if the faces are all closed polygons,
        and the cells closed polyhedra.
        """
        is_valid_face = self.edge_df.groupby("face").apply(_test_valid)
        is_valid = self.upcast_face(is_valid_face)
        if "cell" in self.data_names:
            is_valid_cell = self.edge_df.groupby("cell").apply(_is_closed_cell)
            is_valid = np.logical_and(is_valid, self.upcast_cell(is_valid_cell))
        self.edge_df["is_valid"] = is_valid
        return is_valid

    def get_invalid(self):
        """Returns a mask over self.edge_df for invalid faces
        """
        is_valid = self.get_valid()
        return ~is_valid

    def sanitize(self, trim_borders=False):
        """Removes invalid faces and associated vertices

        If trim_borders is True (defaults to False), there will be a single
        border edge per border face.
        """
        invalid_edges = self.get_invalid()
        if not len(invalid_edges) and trim_borders:
            from ..topology.base_topology import merge_border_edges

            merge_border_edges(self)

        self.remove(invalid_edges, trim_borders)

    def remove(self, edge_out, trim_borders=False):
        """Remove the edges indexed by `edge_out` associated with all
        the cells and faces containing those edges

        If trim_borders is True (defaults to False), there will be a single
        border edge per border face.
        """
        top_level = self.element_names[-1]
        log.info("Removing cells at the %s level", top_level)
        fto_rm = self.edge_df.loc[edge_out, top_level].unique()
        if not fto_rm.shape[0]:
            log.info("Nothing to remove")
            return
        if fto_rm.shape[0] == self.datasets[top_level].shape[0]:
            raise ValueError("sanitize would delete the whole epithlium")

        fto_rm.sort()
        log.info("%d %s level elements will be removed", len(fto_rm), top_level)

        edge_df_ = (
            self.edge_df.set_index(top_level, append=True).swaplevel(0, 1).sort_index()
        )
        to_rm = np.concatenate([edge_df_.loc[c].index.values for c in fto_rm])
        to_rm.sort()
        self.edge_df = self.edge_df.drop(to_rm)

        remaining_verts = np.unique(self.edge_df[["srce", "trgt"]])
        self.vert_df = self.vert_df.loc[remaining_verts]
        if top_level == "face":
            self.face_df = self.face_df.drop(fto_rm)
        elif top_level == "cell":
            remaining_faces = np.unique(self.edge_df["face"])
            self.face_df = self.face_df.loc[remaining_faces]
            self.cell_df = self.cell_df.drop(fto_rm)
        self.reset_index()
        self.reset_topo()
        if trim_borders:
            from ..topology.base_topology import merge_border_edges

            merge_border_edges(self)

    def cut_out(self, bbox, coords=None):
        """Returns the index of edges with at least one vertex outside of the bounding box

        Parameters
        ----------
        bbox : sequence of shape (dim, 2)
             the bounding box as (min, max) pairs for each coordinates.
        coords : list of str of len dim, default None
             the coords corresponding to the bbox.
        """
        if coords is None:
            coords = self.coords
        outs = pd.DataFrame(index=self.edge_df.index, columns=coords)
        for c, bounds in zip(coords, bbox):
            out_vert_ = (self.vert_df[c] < bounds[0]) | (self.vert_df[c] > bounds[1])
            outs[c] = self.upcast_srce(out_vert_) | self.upcast_trgt(out_vert_)

        edge_out = outs.sum(axis=1).astype(np.bool)
        return self.edge_df[edge_out].index

    def set_bbox(self, margin=0.0):
        """Sets the attribute `bbox` with pairs of values bellow
        and above the min and max of the vert coords, with a margin.
        """
        self.bbox = np.array(
            [
                [self.vert_df[c].min() - margin, self.vert_df[c].max() + margin]
                for c in self.coords
            ]
        )

    def reset_index(self):
        """Resets the datasets  to have continuous indices
        """
        log.debug("reseting index for %s", self.identifier)
        self.topo_changed = True

        # remove disconnected vertices and faces
        self.vert_df = self.vert_df.reindex(
            set(self.edge_df.srce).union(self.edge_df.trgt)
        )
        self.face_df = self.face_df.reindex(set(self.edge_df.face))

        new_vidx = pd.Series(np.arange(self.vert_df.shape[0]), index=self.vert_df.index)

        self.edge_df["srce"] = new_vidx.reindex(self.edge_df["srce"]).values.astype(int)
        self.edge_df["trgt"] = new_vidx.reindex(self.edge_df["trgt"]).values.astype(int)

        new_fidx = pd.Series(np.arange(self.face_df.shape[0]), index=self.face_df.index)

        self.edge_df["face"] = new_fidx.loc[self.edge_df["face"]].values.astype(int)

        self.vert_df.reset_index(drop=True, inplace=True)
        self.vert_df.index.name = "vert"

        self.face_df.reset_index(drop=True, inplace=True)
        self.face_df.index.name = "face"

        if "cell" in self.data_names:
            new_cidx = pd.Series(
                np.arange(self.cell_df.shape[0]), index=self.cell_df.index
            )
            self.edge_df["cell"] = new_cidx.loc[self.edge_df["cell"]].values
            self.cell_df.reset_index(drop=True, inplace=True)
            self.cell_df.index.name = "cell"

        self.edge_df.reset_index(drop=True, inplace=True)
        self.edge_df.index.name = "edge"

    def triangular_mesh(self, coords=None, return_mask=False):
        """
        Return a triangulation of an epithelial sheet (2D in a 3D space),
        with added edges between face barycenters and junction vertices.

        Parameters
        ----------
        coords : list of str:
          pair of coordinates corresponding to column names
          for self.face_df and self.vert_df

        return_mask : bool, optional, default True
          if True, returns `face_mask`

        Returns
        -------
        vertices : (self.Nf+self.Nv, 3) ndarray
           all the vertices' coordinates
        triangles : (self.Ne, 3) ndarray of ints
           triple of the vertices' indexes forming
           the triangular elements. For each junction edge, this is simply
           the index (srce, trgt, face). This is correctly oriented.
        face_mask: (self.Nf + self.Nv,) mask with 1 iff the vertex corresponds
           to a face center
        """
        if coords is None:
            coords = self.coords

        vertices = np.concatenate((self.face_df[coords], self.vert_df[coords]), axis=0)

        # edge indices as (Nf + Nv) * 3 array
        triangles = self.edge_df[["srce", "trgt", "face"]].values
        # The src, trgt, face triangle is correctly oriented
        # both vert_idx cols are shifted by Nf
        triangles[:, :2] += self.Nf

        # Ensure returned arrays are C-contiguous
        if not vertices.data.c_contiguous:
            vertices = np.array(vertices, order="C")

        if not triangles.data.c_contiguous:
            triangles = np.array(triangles, order="C")

        if not return_mask:
            return vertices, triangles

        face_mask = np.arange(self.Nf + self.Nv) < self.Nf
        return vertices, triangles, face_mask

    def vertex_mesh(self, coords, vertex_normals=True):
        """ Returns the vertex coordinates and a list of vertex indices
        for each face of the tissue.
        If `vertex_normals` is True, also returns the normals of each vertex
        (set as the average of the vertex' edges), suitable for .OBJ export
        """
        vertices = self.vert_df[coords]
        faces = self.edge_df.groupby("face").apply(_ordered_vert_idxs)
        faces = faces.dropna()

        if vertex_normals:
            normals = (
                self.edge_df.groupby("srce")[self.ncoords].mean()
                + self.edge_df.groupby("trgt")[self.ncoords].mean()
            ) / 2.0
            return vertices.values, faces.values, normals.values
        return vertices.values, faces.values

    def validate_closed_cells(self):
        is_closed = self.edge_df.groupby("cell").apply(_is_closed_cell)
        return is_closed

    def get_opposite_faces(self):
        """Populates the 'opposite' column of self.face_df with the index of
        the opposite face or -1 if the face has no opposite.

        """
        face_v = self.edge_df.groupby("face").apply(lambda df: frozenset(df["srce"]))
        face_v2 = pd.Series(data=face_v.index, index=face_v.values)
        grouped = face_v2.groupby(level=0)
        cardinal = grouped.apply(len)
        if cardinal.max() > 2:
            raise ValueError(
                "Invalid topology, faces have more than one neighbor: {}".format(
                    list(face_v2[cardinal > 2].index)
                )
            )
        self.face_df["opposite"] = -1

        face_pairs = (
            face_v2[cardinal == 2]
            .groupby(level=0)
            .apply(lambda df: dict(enumerate(df)))
            .values.reshape((-1, 2))
        )
        if not face_pairs.shape[0]:
            return
        self.face_df.loc[face_pairs[:, 0], "opposite"] = face_pairs[:, 1]
        self.face_df.loc[face_pairs[:, 1], "opposite"] = face_pairs[:, 0]


def get_opposite_faces(eptm):
    warnings.warn("Deprecated, use `eptm.get_opposite_faces()` instead")
    eptm.get_opposite_faces()


def _ordered_edges(face_edges):
    """Returns "srce", "trgt" and "face" indices
    organized clockwise for each face.

    Parameters
    ----------
    face_edges: `pd.DataFrame`
        exerpt of an edge_df for a single face

    Returns
    -------
    edges: list of 3 ints
        srce, trgt, face indices, ordered
    """

    srces, trgts, faces = face_edges[["srce", "trgt", "face"]].values.T
    srce, trgt, face_ = srces[0], trgts[0], faces[0]
    edges = [[srce, trgt, face_]]
    for face_ in faces[1:]:
        srce, trgt = trgt, trgts[srces == trgt][0]
        edges.append([srce, trgt, face_])
    return edges


def _ordered_vert_idxs(face):
    try:
        return [idxs[0] for idxs in _ordered_edges(face)]
    except IndexError:
        return np.nan


def get_next_edges(sheet):
    """
    returns a pd.Series with the index of the next
    edge for each edge
    """
    next_e = sheet.edge_df.groupby("face").apply(_next_edge)
    next_e.index = next_e.index.droplevel("face")
    return next_e.sort_index()


def get_prev_edges(sheet):
    """
    returns a pd.Series with the index of the next
    edge for each edge
    """
    prev_e = sheet.edge_df.groupby("face").apply(_prev_edge)
    prev_e.index = prev_e.index.droplevel("face")
    return prev_e.sort_index()


def _next_edge(edf):

    edf["edge"] = edf.index
    next_edge = edf.set_index("srce", append=False).loc[edf["trgt"], "edge"].values
    return pd.Series(index=edf.index, data=next_edge)


def _prev_edge(edf):

    edf["edge"] = edf.index
    next_edge = edf.set_index("trgt", append=False).loc[edf["srce"], "edge"].values
    return pd.Series(index=edf.index, data=next_edge)


def _is_closed_cell(e_df):
    edges = e_df[["srce", "trgt"]]
    for _, (srce, trgt) in edges.iterrows():
        if edges[(edges["srce"] == trgt) & (edges["trgt"] == srce)].index.size != 1:
            return False
    return True


def _test_invalid(face):
    """ Returns True if the source and target sets of the faces polygon
    are different or if the face polygon is not closed
    """

    s1 = set(face["srce"])
    s2 = set(face["trgt"])
    if s1 != s2:
        return True
    ordered = np.array(_ordered_edges(face))
    if not np.all(ordered[:, 0] == np.roll(ordered[:, 1], 1)):
        return True
    return False


def _test_valid(face):
    """ Returns true iff all sources are also targets for the faces polygon
    """
    return np.logical_not(_test_invalid(face))



import logging
import numpy as np
from functools import wraps

import warnings


from .base_topology import add_vert, collapse_edge, close_face, remove_face
from .base_topology import split_vert as base_split_vert
from tyssue.utils.decorators import do_undo, validate


logger = logging.getLogger(name=__name__)
MAX_ITER = 100


def split_vert(
    sheet, vert, face=None, multiplier=1.5, reindex=True, recenter=False, epsilon=None
):
    """Splits a vertex towards the center of the face.

    This operation removes the  face `face` from the neighborhood of the vertex.
    """
    # Get the value for the length of the new edge
    if epsilon is None:
        epsilon = sheet.settings.get("threshold_length", 0.1) * multiplier
    else:
        warnings.warn(
            "The epsilon argument is deprecated and will be removed in a future version. "
            "The length of the new edge should be set by "
            "`sheet.settings['threshold_length]*multiplier` "
        )

    if face is None:
        face = np.random.choice(sheet.edge_df[sheet.edge_df["srce"] == vert]["face"])

    face_edges = sheet.edge_df.query(f"face == {face}")
    prev_v, = face_edges[face_edges["trgt"] == vert]["srce"]
    next_v, = face_edges[face_edges["srce"] == vert]["trgt"]
    connected = sheet.edge_df[
        sheet.edge_df["trgt"].isin((next_v, prev_v))
        | sheet.edge_df["srce"].isin((next_v, prev_v))
    ]

    base_split_vert(sheet, vert, face, connected, epsilon, recenter)
    for face_ in connected["face"]:
        close_face(sheet, face_)

    if reindex:
        sheet.reset_index()
        sheet.reset_topo()

    return 0


def type1_transition(
    sheet, edge01, *, epsilon=None, remove_tri_faces=True, multiplier=1.5
):
    """Performs a type 1 transition around the edge edge01

    See ../../doc/illus/t1_transition.png for a sketch of the definition
    of the vertices and cells letterings
    See Finegan et al. for a description of the algotithm https://doi.org/10.1101/704932


    Parameters
    ----------
    sheet : a `Sheet` instance
    edge_01 : int
       index of the edge around which the transition takes place
    epsilon : float, optional, deprecated
       default 0.1, the initial length of the new edge, in case "threshold_length"
       is not in the sheet.settings
    remove_tri_faces : bool, optional
       if True (the default), will remove triangular cells
       after the T1 transition is performed
    multiplier : float, optional
       default 1.5, the multiplier to the threshold length, so that the
       length of the new edge is set to multiplier * threshold_length


    """

    srce, trgt, face = sheet.edge_df.loc[edge01, ["srce", "trgt", "face"]]

    vert = min(srce, trgt)  # find the vertex that won't be reindexed
    ret_code = collapse_edge(sheet, edge01, reindex=True)
    if ret_code != 0:
        warnings.warn(f"Collapse of edge {edge01} failed")
        return ret_code

    split_vert(
        sheet,
        vert,
        face,
        multiplier=multiplier,
        reindex=True,
        recenter=True,
        epsilon=epsilon,
    )

    if not remove_tri_faces:
        return 0
    # Type 1 transitions might create 3 or 2 sided cells, we remove those
    tri_faces = sheet.face_df[sheet.face_df["num_sides"] < 4].index
    i = 0
    while len(tri_faces):
        remove_face(sheet, tri_faces[0])
        tri_faces = sheet.face_df[sheet.face_df["num_sides"] < 4].index
        i += 1
        if i > MAX_ITER:
            raise RecursionError
    return 0


def cell_division(sheet, mother, geom, angle=None):

    if not sheet.face_df.loc[mother, "is_alive"]:
        logger.warning("Cell %s is not alive and cannot devide", mother)
        return
    edge_a, edge_b = get_division_edges(sheet, mother, geom, angle=angle, axis="x")
    if edge_a is None:
        return

    vert_a, new_edge_a, new_opp_edge_a = add_vert(sheet, edge_a)
    vert_b, new_edge_b, new_opp_edge_b = add_vert(sheet, edge_b)
    sheet.vert_df.index.name = "vert"
    daughter = face_division(sheet, mother, vert_a, vert_b)
    return daughter


def get_division_edges(sheet, mother, geom, angle=None, axis="x"):

    if angle is None:
        angle = np.random.random() * np.pi

    m_data = sheet.edge_df[sheet.edge_df["face"] == mother]
    # if angle == 0:
    #     face_pos = sheet.face_df.loc[mother, sheet.coords]
    #     rot_pos = sheet.vert_df[sheet.coords].copy()
    #     for c in sheet.coords:
    #         rot_pos.loc[:, c] = rot_pos[c] - face_pos[c]
    # else:
    rot_pos = geom.face_projected_pos(sheet, mother, psi=angle)

    srce_pos = rot_pos.loc[m_data["srce"], axis]
    srce_pos.index = m_data.index
    trgt_pos = rot_pos.loc[m_data["trgt"], axis]
    trgt_pos.index = m_data.index
    try:
        edge_a = m_data[(srce_pos < 0) & (trgt_pos >= 0)].index[0]
        edge_b = m_data[(srce_pos >= 0) & (trgt_pos < 0)].index[0]
    except IndexError:
        print("Failed")
        logger.error("Division of Cell {} failed".format(mother))
        return None, None
    return edge_a, edge_b


def face_division(sheet, mother, vert_a, vert_b):
    """
    Divides the face associated with edges
    indexed by `edge_a` and `edge_b`, splitting it
    in the middle of those edes.
    """
    # mother = sheet.edge_df.loc[edge_a, 'face']

    face_cols = sheet.face_df.loc[mother]
    sheet.face_df = sheet.face_df.append(face_cols, ignore_index=True)
    sheet.face_df.index.name = "face"
    daughter = int(sheet.face_df.index[-1])

    edge_cols = sheet.edge_df[sheet.edge_df["face"] == mother].iloc[0]
    sheet.edge_df = sheet.edge_df.append(edge_cols, ignore_index=True)
    new_edge_m = sheet.edge_df.index[-1]
    sheet.edge_df.loc[new_edge_m, "srce"] = vert_b
    sheet.edge_df.loc[new_edge_m, "trgt"] = vert_a

    sheet.edge_df = sheet.edge_df.append(edge_cols, ignore_index=True)
    new_edge_d = sheet.edge_df.index[-1]
    sheet.edge_df.loc[new_edge_d, "srce"] = vert_a
    sheet.edge_df.loc[new_edge_d, "trgt"] = vert_b

    # ## Discover daughter edges
    m_data = sheet.edge_df[sheet.edge_df["face"] == mother]
    daughter_edges = [new_edge_d]
    srce, trgt = vert_a, vert_b
    srces, trgts = m_data[["srce", "trgt"]].values.T

    while trgt != vert_a:
        srce, trgt = trgt, trgts[srces == trgt][0]
        daughter_edges.append(
            m_data[(m_data["srce"] == srce) & (m_data["trgt"] == trgt)].index[0]
        )
    sheet.edge_df.loc[daughter_edges, "face"] = daughter
    sheet.edge_df.index.name = "edge"
    sheet.reset_topo()
    return daughter


def resolve_t1s(sheet, geom, model, solver, max_iter=60):

    l_th = sheet.settings["threshold_length"]
    i = 0
    while sheet.edge_df.length.min() < l_th:

        for edge in (
            sheet.edge_df[sheet.edge_df.length < l_th].sort_values("length").index
        ):
            try:
                type1_transition(sheet, edge)
            except KeyError:
                continue
            sheet.reset_index()
            sheet.reset_topo()
            geom.update_all(sheet)
        solver.find_energy_min(sheet, geom, model)
        i += 1
        if i > max_iter:
            break


def _cast_to_int(df_value):

    if len(df_value) == 1:
        return int(df_value)
    elif len(df_value) == 0:
        return -1
    else:
        raise ValueError("Trying to retrieve an integer from a more than length 1 df ")



"""Utilities to generate point clouds

The positions of the points are generated along the architecture
of the epithelium.
"""
from collections import abc

import numpy as np
import pandas as pd
from ..config.subdiv import bulk_spec


class EdgeSubdiv:
    """
    Container class to ease discretisation along the edges

    """

    def __init__(self, edge_df, **kwargs):
        """Creates an indexer and an offset array  to ease
        discretisation along the edges.

        Parameters
        ----------
        edge_df: pd.DataFrame,

        Keyword parameters
        ------------------
        density: number of points per edge


        Attributes
        ----------

        upcaster: np.ndarray, shape (Np,)
          edge indices repeated to match the lookup table
        offset: np.ndarray, shape (Np,)
          piecewise linear offset along the edges, such that
          ::math:M_{ij}^n = offset[n]*r_{ij}:

        """

        self.edge_df = edge_df.copy()
        self.n_edges = self.edge_df.shape[0]
        self.specs = bulk_spec()
        self.specs.update(**kwargs)
        self.unit_length = edge_df.length.mean()

        if "density" not in edge_df:
            self.edge_df["density"] = self.specs["density"]
        self.n_points = 0
        self.points = None
        self.offset_lut = None
        self.update_all()

    def update_all(self):
        self.update_offset_lut()
        self.update_particles()
        self.update_upcaster()
        self.update_offset()

    @staticmethod
    def from_eptm_edges(eptm, **kwargs):
        """Creates an EdgeSubdiv instance and computes the point
        grid allong the edges from the source vertex to its target.
        Returns
        -------
        subdiv: a :class:`EdgeSubdiv` instance

        """

        subdiv = EdgeSubdiv(eptm.edge_df[["length", "density"]], **kwargs)
        srce_pos = eptm.upcast_srce(eptm.vert_df[eptm.coords])
        r_ij = eptm.edge_df[eptm.dcoords]
        subdiv.edge_point_cloud(srce_pos, r_ij)
        return subdiv

    @staticmethod
    def _offset_lut_(num):
        return np.arange(0.5, num + 0.5) / num

    def update_offset_lut(self, offset_lut=None):
        """
        Updates the density lookup table function.

        The `offset_lut` can be any function
        with a single `num` argument


        Parameters
        ----------
        offset_lut: function, default None,
          edge-wise function of the number of points
          giving the offset positions
        Default is a shifted regular grid:
        `np.arange(0.5, num+0.5) / num`

        """
        if offset_lut is None:
            self.offset_lut = self._offset_lut_
        else:
            self.offset_lut = offset_lut

    def update_particles(self):
        """
        * Updates the number of particles per edge from edges length
        and density values:
        `num_particles = length * density`
        * Also updates the self.points df
        """
        self.edge_df["norm_length"] = self.edge_df["length"] / self.unit_length
        points_per_edges = np.round(self.edge_df.eval("norm_length * density")).astype(
            np.int
        )
        self.edge_df["num_particles"] = points_per_edges
        self.n_points = points_per_edges.sum()
        self.points = pd.DataFrame(
            np.zeros((self.n_points, 2)), columns=["upcaster", "offset"]
        )

    def update_upcaster(self):
        """
        resets the 'upcaster' column of self.points,

        'upcaster' indexes over self.edge_df repeated to
        upcast data from the edge df to the points df
        """
        self.points["upcaster"] = np.repeat(
            np.arange(self.edge_df.shape[0]), self.edge_df["num_particles"]
        )

    def update_offset(self):
        self.points["offset"] = np.concatenate(
            [self.offset_lut(num=ns) for ns in self.edge_df["num_particles"]]
        )

    def validate(self):

        if not self.points["upcaster"].max() + 1 == self.n_edges:
            return False
        if not self.points["upcaster"].shape[0] == self.n_points:
            return False
        if not self.points["offset"].shape()[0] == self.n_points:
            return False
        return True

    def upcast(self, df):
        if isinstance(df, str) and df in self.edge_df:
            return self.edge_df.loc[self.points["upcaster"], df]
        elif (
            isinstance(df, abc.Iterable)
            and isinstance(df[0], str)
            and set(df).issubset(self.edge_df.columns)
        ):
            return self.edge_df.loc[self.points["upcaster"], df]
        elif hasattr(df, "loc"):
            return df.loc[self.points["upcaster"]]
        else:
            raise ValueError(
                """
Argument df should be a column name or a sequence of column names
or a Series or Dataframe indexed like self.edge_df
                """
            )

    def edge_point_cloud(
        self,
        srce_pos,
        r_ij,
        offset_modulation=None,
        modulation_kwargs=None,
        coords=["x", "y", "z"],
        dcoords=["dx", "dy", "dz"],
    ):
        """Generates a point cloud along the edges of the epithelium.

        if a offset_modulation function is provided, it is used to
        transform the offsets

        Parameters
        ----------
        srce_pos: DataFrame of shape (self.Ne, ndim)
          with the origins of the points for each edge
          (usually the edge upcasted source vertex)

        r_ij: DataFrame of shape (self.Ne, ndim)
          the edge vector coordiantes

        offset_modulation: function of self returning
          an array with shape (self.Np,) containing
          the modified offsets. self.points['offset']
          is used by default.
        modulation_kwargs: keyword arguments to the modulation
          function

        Returns
        -------
        points: (Np, 3) pd.DataFrame with the points positions
        """
        for u, du in zip(coords, dcoords):
            self.edge_df[u] = srce_pos[u]
            self.edge_df["d" + u] = r_ij[du]
        cols = coords + dcoords
        upcast = self.edge_df.loc[self.points["upcaster"], cols]
        if offset_modulation is None:
            upcast["offset"] = self.points["offset"].values
        else:
            upcast["offset"] = offset_modulation(self, **modulation_kwargs)
        for c in coords:
            self.points[c] = upcast.eval("{} + offset * {}".format(c, "d" + c)).values
        if self.specs["noise"] > 0.0:
            self.points[coords] += np.random.normal(
                scale=self.specs["noise"], size=(self.n_points, 3)
            )
        return self.points[coords]


def get_edge_bases(eptm, base=("face", "srce", "trgt")):

    edge_upcast_pos = {
        element: eptm.upcast_cols(element, eptm.coords) for element in base
    }
    origin = base[0]
    edge_bases = {}
    for vertex in base[1:]:
        df = pd.DataFrame(
            0, columns=eptm.coords + eptm.dcoords + ["length"], index=eptm.edge_df.index
        )
        df[eptm.dcoords] = (
            edge_upcast_pos[vertex].values - edge_upcast_pos[origin].values
        )

        df["length"] = np.linalg.norm(df[eptm.dcoords].values, axis=1)
        df[eptm.coords] = edge_upcast_pos[origin].values
        edge_bases["{}_{}".format(origin, vertex)] = df.copy()
    return edge_bases


class FaceGrid:
    def __init__(self, edges_df, base, **kwargs):

        self.origin = base[0]
        self.base = ["{}_{}".format(base[0], other) for other in base[1:]]

        self.specs = bulk_spec()
        self.specs.update(kwargs)
        e_specs = kwargs
        self.subdivs = {key: EdgeSubdiv(edges_df[key], **e_specs) for key in self.base}
        self.n_points = np.product(
            [
                subdiv.edge_df["num_particles"].values
                for subdiv in self.subdivs.values()
            ],
            axis=0,
        ).sum()
        self.dim = len(self.subdivs)
        self.up_cols = ["up_{}".format(key) for key in self.base]
        self.of_cols = ["of_{}".format(key) for key in self.base]
        self.points = None

    def update_grid(self):
        upcasters = {}
        for key, subdiv in self.subdivs.items():
            upcasters["up_" + key] = subdiv.points["upcaster"]
            upcasters["of_" + key] = subdiv.points["offset"]
        upcasters = pd.DataFrame.from_dict(upcasters)
        points = {}
        u_axis = "up_{}".format(self.base[0])
        upcasters.set_index(u_axis, drop=False, inplace=True)
        for cols in (self.of_cols, self.up_cols):
            df = upcasters.groupby(level=u_axis).apply(_local_grid, *cols)
            points.update({col: df[col].values for col in cols})
        self.points = pd.DataFrame.from_dict(points)

    def face_point_cloud(self, coords=["x", "y", "z"], dcoords=["dx", "dy", "dz"]):
        upcast = {}
        offsets = self.points[self.of_cols]
        for key, subdiv in self.subdivs.items():
            upcast[key] = subdiv.edge_df.loc[
                self.points["up_{}".format(key)], coords + dcoords + ["length"]
            ].copy()
            upcast[key].reset_index(inplace=True)
            upcast[key]["offset"] = offsets["of_{}".format(key)].values

        for u, du in zip(coords, dcoords):
            self.points[u] = upcast[self.base[0]].eval(
                "{} + offset * {}".format(u, du)
            ).values + np.sum(
                [
                    upcast[other].eval("offset * {}".format(du)).values
                    for other in self.base[1:]
                ],
                axis=0,
            )

        in_out = np.zeros(self.points.shape[0], dtype=bool)
        for other in self.base[1:]:
            xx = upcast[self.base[0]]["offset"].values
            yy = upcast[other]["offset"].values
            in_out += (xx + yy) < 1.0

        self.points = self.points[in_out]
        return self.points[coords]


def _local_grid(df, *cols):

    grid = np.meshgrid(*(df[col] for col in cols))
    out = pd.DataFrame.from_dict({col: mm.ravel() for col, mm in zip(cols, grid)})
    return out



import warnings
import numpy as np
from copy import deepcopy

from .effectors import dimensionalize as dimensionalize
from .effectors import normalize as normalize

from ..utils import to_nd


def model_factory(effectors, ref_effector=None):
    """Produces a Model class with the provided effectors.

    Parameters
    ----------
    effectors : list of :class:`.effectors.AbstractEffectors` classes.
    ref_effector : optional, default None
        if passed, will be used for normalization,
        by default, the last effector in the list is used

    Returns
    -------
    NewModel : a Model derived class with compute_enregy and compute_gradient
      methods

    """
    if ref_effector is None:
        ref_effector = effectors[-1]

    class NewModel:

        labels = []
        specs = {
            "cell": {},
            "face": {},
            "edge": {},
            "vert": {},
            "settings": {"nrj_norm_factor": 1.0},
        }
        for f in effectors:
            labels.append(f.label)
            for k in specs:
                specs[k].update(f.specs.get(k, {}))

        @staticmethod
        def dimensionalize(nondim_specs):
            dim_specs = deepcopy(nondim_specs)
            for effector in effectors:
                if effector == ref_effector:
                    continue
                dimensionalize(nondim_specs, dim_specs, effector, ref_effector)

            ref_nrj = ref_effector.get_nrj_norm(dim_specs)
            dim_specs["settings"]["nrj_norm_factor"] = ref_nrj
            return dim_specs

        @classmethod
        def dimentionalize(cls, nondim_specs):
            warnings.warn(
                """This badly worded method is deprecated,
 use dimensionalize instead"""
            )
            return cls.dimensionalize(nondim_specs)

        @staticmethod
        def normalize(dim_specs):
            nondim_specs = deepcopy(dim_specs)
            for effector in effectors:
                normalize(dim_specs, nondim_specs, effector, ref_effector)

        @staticmethod
        def compute_energy(eptm, full_output=False):
            energies = [f.energy(eptm) for f in effectors]
            norm_factor = eptm.specs["settings"].get("nrj_norm_factor", 1)
            if full_output:
                return [E / norm_factor for E in energies]

            return sum(E.sum() for E in energies) / norm_factor

        @staticmethod
        def compute_gradient(eptm, components=False):
            norm_factor = eptm.specs["settings"].get("nrj_norm_factor", 1)
            if not eptm.ucoords[0] in eptm.edge_df.columns:
                warnings.warn(
                    "setting ucoords in grad computation," "please fix your specs"
                )
                for uc in eptm.ucoords:
                    eptm.edge_df[uc] = 0.0

            eptm.edge_df[eptm.ucoords] = eptm.edge_df[eptm.dcoords] / to_nd(
                eptm.edge_df["length"], eptm.dim
            )

            grads = [f.gradient(eptm) for f in effectors]
            if components:
                return grads

            grad_s, grad_t, grad_v = None, None, None

            srce_grads = [g[0] for g in grads if g[0].shape[0] == eptm.Ne]
            if srce_grads:
                grad_s = eptm.sum_srce(sum(srce_grads))
            trgt_grads = [
                g[1] for g in grads if (g[1] is not None) and (g[1].shape[0] == eptm.Ne)
            ]
            if trgt_grads:
                grad_t = eptm.sum_trgt(sum(trgt_grads))
            vert_grads = [g[0] for g in grads if g[0].shape[0] == eptm.Nv]
            if vert_grads:
                grad_v = sum(vert_grads)

            grad_i = sum([g for g in (grad_s, grad_t, grad_v) if g is not None])
            return grad_i / norm_factor

    return NewModel



import numpy as np

try:
    from .cpp import c_collisions
except ImportError:
    print(
        "collision solver could not be imported "
        "You may need to install CGAL and re-install tyssue"
    )
    c_collisions = None


def self_intersections(sheet):
    """Checks for self collisions for the sheet

    Parameters
    ----------
    sheet : a :class:`Sheet` object
        This object must have a `triangular_mesh` method returning a
        valid triangular mesh.

    Returns
    -------
    edge_pairs: np.ndarray of indices
         Array of shape (n_intersections, 2) with the indices of the
         pairs of intersecting edges
    """
    faces, vertices = sheet.triangular_mesh(sheet.coords, return_mask=False)
    mesh = c_collisions.sheet_to_surface_mesh(faces, vertices)
    if not c_collisions.does_self_intersect(mesh):
        return np.empty((0, 2), dtype=np.int)
    return np.array(c_collisions.self_intersections(mesh), dtype=np.int)
